# Copyright 2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Federated Bootstrap Aggregation for XGBoost module."""

import json

import numpy as np

from openfl.interface.aggregation_functions.core import AggregationFunction


def get_global_model(iterator, target_round):
    """
    Retrieves the global model for the specific round from an iterator.

    Parameters:
    iterator (iterable): An iterable containing items with 'tags' and 'round' keys.
    target_round (int): The round number for which the global model is to be retrieved.

    Returns:
    np.ndarray: The numpy array representing the global model for the specified round.
    """
    for item in iterator:
        # Items tagged with ('model',) are the global model of that round
        if "tags" in item and item["tags"] == ("model",) and item["round"] == target_round:
            return item["nparray"]
    raise ValueError(f"No item found with tag 'model' and round {target_round}")


def append_trees(global_model, local_trees):
    """
    Appends local trees to the global model.

    Parameters:
    global_model (dict): A dictionary representing the global model.
    local_trees (list): A list of dictionaries representing the local trees to be appended to the
        global model.

    Returns:
    dict: The updated global model with the local trees appended.
    """
    num_global_trees = int(
        global_model["learner"]["gradient_booster"]["model"]["gbtree_model_param"]["num_trees"]
    )
    num_local_trees = len(local_trees)

    global_model["learner"]["gradient_booster"]["model"]["gbtree_model_param"]["num_trees"] = str(
        num_global_trees + num_local_trees
    )
    global_model["learner"]["gradient_booster"]["model"]["iteration_indptr"].append(
        num_global_trees + num_local_trees
    )
    for new_tree in range(num_local_trees):
        local_trees[new_tree]["id"] = num_global_trees + new_tree
        global_model["learner"]["gradient_booster"]["model"]["trees"].append(local_trees[new_tree])
        global_model["learner"]["gradient_booster"]["model"]["tree_info"].append(0)

    return global_model


class FedBaggingXGBoost(AggregationFunction):
    """
    Federated Bootstrap Aggregation for XGBoost.

    This class implements a federated learning aggregation function specifically
    designed for XGBoost models. It aggregates local model updates (trees) from
    multiple collaborators into a global model using a bagging approach.
    """

    def call(self, local_tensors, db_iterator, tensor_name, fl_round, *_):
        """Aggregate tensors.

        Args:
            local_tensors (list[openfl.utilities.LocalTensor]): List of local
                tensors to aggregate.
            db_iterator: iterator over history of all tensors. Columns:
                - 'tensor_name': name of the tensor.
                    Examples for `torch.nn.Module`s: 'conv1.weight','fc2.bias'.
                - 'round': 0-based number of round corresponding to this
                    tensor.
                - 'tags': tuple of tensor tags. Tags that can appear:
                    - 'model' indicates that the tensor is a model parameter.
                    - 'trained' indicates that tensor is a part of a training
                        result.
                        These tensors are passed to the aggregator node after
                        local learning.
                    - 'aggregated' indicates that tensor is a result of
                        aggregation.
                        These tensors are sent to collaborators for the next
                        round.
                    - 'delta' indicates that value is a difference between
                        rounds for a specific tensor.
                    also one of the tags is a collaborator name
                    if it corresponds to a result of a local task.

                - 'nparray': value of the tensor.
            tensor_name: name of the tensor
            fl_round: round number
            tags: tuple of tags for this tensor
        Returns:
            bytearray: aggregated tensor
        """
        global_model = get_global_model(db_iterator, fl_round)

        if (
            isinstance(global_model, np.ndarray) and global_model.size == 0
        ) or global_model is None:
            # if there is no global model, use the first local model as the global model
            for local_tensor in local_tensors:
                local_tree_bytearray = bytearray(local_tensor.tensor.astype(np.uint8).tobytes())
                local_tree_json = json.loads(local_tree_bytearray)

                if (
                    isinstance(global_model, np.ndarray) and global_model.size == 0
                ) or global_model is None:
                    # the first tree becomes the global model
                    global_model = local_tree_json
                else:
                    # append subsequent trees to global model
                    local_model = local_tree_json
                    local_trees = local_model["learner"]["gradient_booster"]["model"]["trees"]
                    global_model = append_trees(global_model, local_trees)
        else:
            global_model_bytearray = bytearray(global_model.astype(np.uint8).tobytes())
            # convert the global model to a dictionary
            global_model = json.loads(global_model_bytearray)

            for local_tensor in local_tensors:
                # append trees to global model
                local_tree_bytearray = bytearray(local_tensor.tensor.astype(np.uint8).tobytes())
                local_trees = json.loads(local_tree_bytearray)
                global_model = append_trees(global_model, local_trees)

        global_model_json = json.dumps(global_model)
        global_model_bytes = global_model_json.encode("utf-8")

        global_model_float32_array = np.frombuffer(global_model_bytes, dtype=np.uint8).astype(
            np.float32
        )

        return global_model_float32_array



# Copyright (C) 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
from openfl.federated.data.loader import DataLoader
from sklearn.datasets import load_iris
import pandas as pd


class IRISInMemory(DataLoader):
    """Data Loader for IRIS Dataset."""

    def __init__(self, batch_size, data_path, **kwargs):
        super().__init__(**kwargs)

        # download data
        self._download_raw_data()
        # create shards
        self.data_shard = self.load_data_shard(
            shard_num=int(data_path), **kwargs
        )

    def _download_raw_data(self):
        """
        Downloads the raw Iris dataset and saves it as a CSV file.
        This method loads the Iris dataset using the `load_iris` function from
        the `sklearn.datasets` module. The dataset is then converted to a
        pandas DataFrame and saved as a CSV file named 'client.csv' in the
        './data/' directory.
        Returns:
            None
        """

        iris = load_iris(as_frame=True)
        data = iris['data']
        data.to_csv('./data/client.csv', index=False)

    def _load_data(self):
        """
        Loads data from a CSV file.
        This method reads the data from a CSV file located at './data/client.csv'
        and returns it as a pandas DataFrame.
        Returns:
            pd.DataFrame: The data loaded from the CSV file.
        """

        return pd.read_csv('./data/client.csv')

    def load_data_shard(self, shard_num, collaborator_count, **kwargs):
        """
        Load a specific shard of dataset for a given collaborator.
        Args:
            shard_num (int): The shard number to load.
            collaborator_count (int): The total number of collaborators.
            **kwargs: Additional keyword arguments.
        Returns:
            pandas.DataFrame: The shard of dataset corresponding to the given shard number.
        """

        return self._load_data().iloc[shard_num-1::collaborator_count]

    def query(self, columns, **kwargs):
        """
        Query the data shard for the specified columns.
        Args:
            columns (list): A list of column names to query from the data shard.
            **kwargs: Additional keyword arguments (currently not used).
        Returns:
            DataFrame: A DataFrame containing the data for the specified columns.
        Raises:
            ValueError: If the columns parameter is not a list.
        """
        if not isinstance(columns, list):
            raise ValueError("Columns parameter must be a list")
        return self.data_shard[columns]

    def get_feature_shape(self):
        """
        This function is not required and is kept for compatibility.

        Returns:
            None
        """
        pass



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from openfl.federated import KerasDataLoader
from .mnist_utils import load_mnist_shard


class KerasMNISTInMemory(KerasDataLoader):
    """Data Loader for MNIST Dataset."""

    def __init__(self, data_path, batch_size, **kwargs):
        """
        Initialize.

        Args:
            data_path: File path for the dataset
            batch_size (int): The batch size for the data loader
            **kwargs: Additional arguments, passed to super init and load_mnist_shard
        """
        super().__init__(batch_size, **kwargs)

        # TODO: We should be downloading the dataset shard into a directory
        # TODO: There needs to be a method to ask how many collaborators and
        #  what index/rank is this collaborator.
        # Then we have a way to automatically shard based on rank and size of
        # collaborator list.
        try:
            int(data_path)
        except:
            raise ValueError(
                "Expected `%s` to be representable as `int`, as it refers to the data shard " +
                "number used by the collaborator.",
                data_path
            )

        _, num_classes, X_train, y_train, X_valid, y_valid = load_mnist_shard(
            shard_num=int(data_path), **kwargs
        )

        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid

        self.num_classes = num_classes



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""SKCPipeline module."""

import copy as co
import gzip as gz

import numpy as np
from sklearn import cluster

from openfl.pipelines.pipeline import TransformationPipeline, Transformer


class SparsityTransformer(Transformer):
    """A transformer class to sparsify input data.

    Attributes:
        p (float): The sparsity ratio.
        lossy (bool): A flag indicating if the transformation is lossy.
    """

    def __init__(self, p=0.01):
        """Initialize.

        Args:
            p (float, optional): The sparsity ratio. Defaults to 0.01.
        """
        self.lossy = True
        self.p = p

    def forward(self, data, **kwargs):
        """Sparsify data and pass over only non-sparsified elements by reducing
        the array size.

        Args:
            data: an numpy array from the model tensor_dict.

        Returns:
            sparse_data: a flattened, sparse representation of the input
                tensor.
            metadata: dictionary to store a list of meta information.
        """
        metadata = {"int_list": list(data.shape)}
        # sparsification
        data = data.astype(np.float32)
        flatten_data = data.flatten()
        n_elements = flatten_data.shape[0]
        k_op = int(np.ceil(n_elements * self.p))
        topk, topk_indices = self._topk_func(flatten_data, k_op)
        sparse_data = np.zeros(flatten_data.shape)
        sparse_data[topk_indices] = topk
        return sparse_data, metadata

    def backward(self, data, metadata, **kwargs):
        """Recover data array with the right shape and numerical type.

        Args:
            data: an numpy array with non-zero values.
            metadata: dictionary to contain information for recovering back
                to original data array.

        Returns:
            recovered_data: an numpy array with original shape.
        """
        data = data.astype(np.float32)
        data_shape = metadata["int_list"]
        recovered_data = data.reshape(data_shape)
        return recovered_data

    @staticmethod
    def _topk_func(x, k):
        """Select top k values.

        Args:
            x: an numpy array to be sorted out for top-k components.
            k: k most maximum values.

        Returns:
            topk_mag: components with top-k values.
            indices: indices of the top-k components.
        """
        # quick sort as default on magnitude
        idx = np.argsort(np.abs(x))
        # sorted order, the right most is the largest magnitude
        length = x.shape[0]
        start_idx = length - k
        # get the top k magnitude
        topk_mag = np.asarray(x[idx[start_idx:]])
        indices = np.asarray(idx[start_idx:])
        if min(topk_mag) - 0 < 10e-8:  # avoid zeros
            topk_mag = topk_mag + 10e-8
        return topk_mag, indices


class KmeansTransformer(Transformer):
    """A transformer class to quantize input data.

    Attributes:
        n_cluster (int): The number of clusters for the K-means.
        lossy (bool): A flag indicating if the transformation is lossy.
    """

    def __init__(self, n_cluster=6):
        """Initialize KmeansTransformer.

        Args:
            n_cluster (int, optional): The number of clusters for the K-means.
                Defaults to 6.
        """
        self.n_cluster = n_cluster
        self.lossy = True

    def forward(self, data, **kwargs):
        """Quantize data into n_cluster levels of values.

        Args:
            data: an flattened numpy array.

        Returns:
            int_data: an numpy array being quantized.
            metadata: dictionary to store a list of meta information.
        """
        # clustering
        data = data.reshape((-1, 1))
        if data.shape[0] >= self.n_cluster:
            k_means = cluster.KMeans(n_clusters=self.n_cluster, n_init=self.n_cluster)
            k_means.fit(data)
            quantized_values = k_means.cluster_centers_.squeeze()
            indices = k_means.labels_
            quant_array = np.choose(indices, quantized_values)
        else:
            quant_array = data
        int_array, int2float_map = self._float_to_int(quant_array)
        metadata = {"int_to_float": int2float_map}
        int_array = int_array.reshape(-1)
        return int_array, metadata

    def backward(self, data, metadata, **kwargs):
        """Recover data array back to the original numerical type.

        Args:
            data: an numpy array with non-zero values.
            metadata: dictionary to contain information for recovering back
                to original data array.

        Returns:
            metadata: dictionary to contain information for recovering back
                to original data array.
            data: an numpy array with original numerical type.
        """
        # convert back to float
        data = co.deepcopy(data)
        int2float_map = metadata["int_to_float"]
        for key in int2float_map:
            indices = data == key
            data[indices] = int2float_map[key]
        return data

    @staticmethod
    def _float_to_int(np_array):
        """Create look-up table for conversion between floating and integer
        types.

        Args:
            np_array: A numpy array.

        Returns:
            int_array: The input numpy float array converted to an integer
                array.
            int_to_float_map: The dictionary mapping integers to floats.
        """
        flatten_array = np_array.reshape(-1)
        unique_value_array = np.unique(flatten_array)
        int_array = np.zeros(flatten_array.shape, dtype=np.int32)
        int_to_float_map = {}
        float_to_int_map = {}
        # create table
        for idx, u_value in enumerate(unique_value_array):
            int_to_float_map.update({idx: u_value})
            float_to_int_map.update({u_value: idx})
            # assign to the integer array
            indices = np.where(flatten_array == u_value)
            int_array[indices] = idx
        int_array = int_array.reshape(np_array.shape)
        return int_array, int_to_float_map


class GZIPTransformer(Transformer):
    """GZIP transformer class for losslessly compressing data.

    Attributes:
        lossy (bool): A flag indicating if the transformation is lossy.
    """

    def __init__(self):
        """Initialize."""
        self.lossy = False

    def forward(self, data, **kwargs):
        """Compress data into bytes.

        Args:
            data: an numpy array with non-zero values.

        Returns:
            compressed_bytes_: The compressed data.
            metadata: An empty dictionary.
        """
        bytes_ = data.astype(np.float32).tobytes()
        compressed_bytes_ = gz.compress(bytes_)
        metadata = {}
        return compressed_bytes_, metadata

    def backward(self, data, metadata, **kwargs):
        """Decompress data into numpy of float32.

        Args:
            data: an numpy array with non-zero values.
            metadata: dictionary to contain information for recovering back
                to original data array.

        Returns:
            data: A numpy array with the original numerical type after
                decompression.
        """
        decompressed_bytes_ = gz.decompress(data)
        data = np.frombuffer(decompressed_bytes_, dtype=np.float32)
        return data


class SKCPipeline(TransformationPipeline):
    """A pipeline class to compress data lossly using sparsity and k-means
    methods.

    Attributes:
        p (float): The sparsity factor.
        n_cluster (int): The number of K-mean clusters.
    """

    def __init__(self, p_sparsity=0.1, n_clusters=6, **kwargs):
        """Initialize a pipeline of transformers.

        Args:
            p_sparsity (float, optional): The sparsity factor. Defaults to 0.1.
            n_clusters (int, optional): The number of K-mean clusters.
                Defaults to 6.
            **kwargs: Additional keyword arguments for the pipeline.

        Returns:
            Data compression transformer pipeline object.
        """
        # instantiate each transformer
        self.p = p_sparsity
        self.n_cluster = n_clusters
        transformers = [
            SparsityTransformer(self.p),
            KmeansTransformer(self.n_cluster),
            GZIPTransformer(),
        ]
        super().__init__(transformers=transformers, **kwargs)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""openfl.experimental.workflow.component package."""

# FIXME: Too much recursion
from openfl.experimental.workflow.component.aggregator import Aggregator
from openfl.experimental.workflow.component.collaborator import Collaborator
from openfl.experimental.workflow.component.director import Director
from openfl.experimental.workflow.component.envoy import Envoy



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Experimental Aggregator module."""

import asyncio
import inspect
import queue
import time
from logging import getLogger
from threading import Event
from typing import Any, Callable, Dict, List, Tuple

import dill

from openfl.experimental.workflow.interface import FLSpec
from openfl.experimental.workflow.runtime import FederatedRuntime
from openfl.experimental.workflow.utilities import aggregator_to_collaborator, checkpoint
from openfl.experimental.workflow.utilities.metaflow_utils import MetaflowInterface

logger = getLogger(__name__)


class Aggregator:
    r"""An Aggregator is the central node in federated learning.

    Args:
        single_col_cert_common_name (str): Common name for single
            collaborator certificate.
        log_metric_callback (optional): Callback for log metric. Defaults
            to None.
        aggregator_uuid (str): Aggregation ID.
        federation_uuid (str): Federation ID.
        authorized_cols (list of str): The list of IDs of enrolled
            collaborators.
        rounds_to_train (int): External loop rounds.
        current_round (int): Current round.
        collaborators_counter (int): counter for the collaborators
        quit_job_sent_to (list): A list of collaborators to whom quit jobs
            will be sent.
        time_to_quit (bool): Time to quit the experiment,
        collaborator_task_results (Event): Event to inform aggregator that
            collaborators have sent the results.
        __collaborator_tasks_queue (Dict[Queue]): queue for each collaborator.
        flow (Any): Flow class.
        name (str): aggregator in string format.
        checkpoint (bool): Whether to save checkpoint or not (default=False).
        private_attrs_callable (Callable): Function for Aggregator private
            attributes.
        (default=None).
        private_attrs (Dict): Private attributes dictionary.
            (default={}).
        connected_collaborators (list): List of connected collaborators
        tasks_sent_to_collaborators (int): count of tasks sent to collaborators.
        stdout_queue (queue.Queue): Queue for stdout.

    Returns:
        None
    """

    def __init__(
        self,
        aggregator_uuid: str,
        federation_uuid: str,
        authorized_cols: List,
        flow: Any,
        rounds_to_train: int = 1,
        checkpoint: bool = False,
        private_attributes_callable: Callable = None,
        private_attributes_kwargs: Dict = None,
        private_attributes: Dict = None,
        single_col_cert_common_name: str = None,
        log_metric_callback: Callable = None,
        **kwargs,
    ) -> None:
        """Initializes the Aggregator.

        Args:
            aggregator_uuid (int): Aggregation ID.
            federation_uuid (str): Federation ID.
            authorized_cols (list of str): The list of IDs of enrolled
                collaborators.
            flow (Any): Flow class.
            rounds_to_train (int, optional): Number of rounds to train.
                Defaults to 1.
            checkpoint (bool): Whether to save checkpoint or not (default=False).
            private_attributes_callable (Callable): Function for Aggregator private
            attributes. (default=None).
            private_attributes (Dict): Private attributes dictionary.
                (default={}).
            single_col_cert_common_name (str, optional): Common name for single
                collaborator certificate. Defaults to None.
            log_metric_callback (optional): Callback for log metric. Defaults
                to None.
            **kwargs: Additional keyword arguments.
        """
        self.single_col_cert_common_name = single_col_cert_common_name
        if self.single_col_cert_common_name is not None:
            self._log_big_warning()
        else:
            # FIXME: "" instead of None is just for protobuf compatibility.
            # Cleaner solution?
            self.single_col_cert_common_name = ""

        self.log_metric_callback = log_metric_callback
        if log_metric_callback is not None:
            self.log_metric = log_metric_callback
            logger.info(f"Using custom log metric: {self.log_metric}")

        self.uuid = aggregator_uuid
        self.federation_uuid = federation_uuid
        self.authorized_cols = authorized_cols

        self.rounds_to_train = rounds_to_train
        self.current_round = 1
        self.collaborators_counter = 0
        self.quit_job_sent_to = []
        self.time_to_quit = False

        # Event to inform aggregator that collaborators have sent the results
        self.collaborator_task_results = Event()
        # A queue for each task
        self.__collaborator_tasks_queue = {collab: queue.Queue() for collab in self.authorized_cols}

        self.flow = flow
        self.checkpoint = checkpoint
        self.flow._foreach_methods = []
        logger.info("MetaflowInterface creation.")
        self.flow._metaflow_interface = MetaflowInterface(self.flow.__class__, "single_process")
        self.flow._run_id = self.flow._metaflow_interface.create_run()
        self.flow.runtime = FederatedRuntime()
        self.name = "aggregator"
        self.flow.runtime.collaborators = self.authorized_cols

        self.__private_attrs_callable = private_attributes_callable
        self.__private_attrs = private_attributes
        self.connected_collaborators = []
        self.tasks_sent_to_collaborators = 0
        self.stdout_queue = queue.Queue()

        if self.__private_attrs_callable is not None:
            logger.info("Initializing aggregator private attributes...")
            self.__initialize_private_attributes(private_attributes_kwargs)

    def __initialize_private_attributes(self, kwargs: Dict) -> None:
        """Call private_attrs_callable function set attributes to
        self.__private_attrs."""
        self.__private_attrs = self.__private_attrs_callable(**kwargs)

    def __set_private_attrs_to_clone(self, clone: Any) -> None:
        """Set private_attrs of Aggregator as attributes of FLSpec clone"""
        if len(self.__private_attrs) > 0:
            for name, attr in self.__private_attrs.items():
                setattr(clone, name, attr)

    def __delete_private_attrs_from_clone(self, clone: Any, replace_str: str = None) -> None:
        """
        Remove aggregator private attributes from FLSpec clone
        before transition from aggregator step to collaborator steps.
        Instead of removing private attributes this method can also replace
        private attributes with a string (required in checkpointing)
        """
        # Update aggregator private attributes by taking latest
        # parameters from clone, then delete attributes from clone.
        if len(self.__private_attrs) > 0:
            for attr_name in self.__private_attrs:
                if hasattr(clone, attr_name):
                    self.__private_attrs.update({attr_name: getattr(clone, attr_name)})
                    if replace_str:
                        setattr(clone, attr_name, replace_str)
                    else:
                        delattr(clone, attr_name)

    def _log_big_warning(self) -> None:
        """Warn user about single collaborator cert mode."""
        logger.warning(
            f"\n{the_dragon}\nYOU ARE RUNNING IN SINGLE COLLABORATOR CERT MODE! THIS IS"
            f" NOT PROPER PKI AND "
            f"SHOULD ONLY BE USED IN DEVELOPMENT SETTINGS!!!! YE HAVE BEEN"
            f" WARNED!!!"
        )

    @staticmethod
    def _get_sleep_time() -> int:
        """Sleep 10 seconds.

        Returns:
            sleep_time: int
        """
        return 10

    async def run_flow(self) -> FLSpec:
        """
        Start the execution and run flow until completion.
        Returns the updated flow to the user.

        Returns:
            flow (FLSpec): Updated instance.
        """
        # Start function will be the first step if any flow
        f_name = "start"
        # Creating a clones from the flow object
        FLSpec._reset_clones()
        FLSpec._create_clones(self.flow, self.flow.runtime.collaborators)

        logger.info(f"Starting round {self.current_round}...")
        while True:
            next_step = self.do_task(f_name)

            if self.time_to_quit:
                logger.info("Experiment Completed.")
                break

            # Prepare queue for collaborator task, with clones
            for k, v in self.__collaborator_tasks_queue.items():
                if k in self.selected_collaborators:
                    v.put((next_step, self.clones_dict[k]))
                else:
                    logger.info(f"Tasks will not be sent to {k}")

            while not self.collaborator_task_results.is_set():
                len_sel_collabs = len(self.selected_collaborators)
                len_connected_collabs = len(self.connected_collaborators)
                if len_connected_collabs < len_sel_collabs:
                    # Waiting for collaborators to connect.
                    logger.info(
                        "Waiting for "
                        + f"{len_sel_collabs - len_connected_collabs}/{len_sel_collabs}"
                        + " collaborators to connect..."
                    )
                elif self.tasks_sent_to_collaborators != len_sel_collabs:
                    logger.info(
                        "Waiting for "
                        + f"{len_sel_collabs - self.tasks_sent_to_collaborators}/{len_sel_collabs}"
                        + " to make requests for tasks..."
                    )
                else:
                    # Waiting for selected collaborators to send the results.
                    logger.info(
                        "Waiting for "
                        + f"{len_sel_collabs - self.collaborators_counter}/{len_sel_collabs}"
                        + " collaborators to send results..."
                    )
                await asyncio.sleep(Aggregator._get_sleep_time())

            self.collaborator_task_results.clear()
            f_name = self.next_step
            if hasattr(self, "instance_snapshot"):
                self.flow.restore_instance_snapshot(self.flow, list(self.instance_snapshot))
                delattr(self, "instance_snapshot")

        return self.flow

    def call_checkpoint(
        self, name: str, ctx: Any, f: Callable, stream_buffer: bytes = None
    ) -> None:
        """
        Perform checkpoint task and log stdout

        Args:
            name (str): name of the caller
            ctx (FLSpec / bytes): Collaborator FLSpec object for which
                checkpoint is to be performed.
            f (Callable / bytes): Collaborator Step (Function) which is to be
                checkpointed.
            stream_buffer (bytes): Captured object for output and error
                (default=None).
        """
        # Perform checkpoint if enabled
        if self.checkpoint:
            if not isinstance(ctx, FLSpec):
                ctx = dill.loads(ctx)
                # Update metaflow interface object
                ctx._metaflow_interface = self.flow._metaflow_interface
            # Deserialize objects if passed in serialized form
            if not isinstance(f, Callable):
                f = dill.loads(f)
            if stream_buffer and isinstance(stream_buffer, bytes):
                f.__func__._stream_buffer = dill.loads(stream_buffer)

            stdout = checkpoint(ctx, f)
            # Retrieve and log stdout
            stream_info = {
                "stdout_origin": name,
                "task_name": f.__name__,
                "stdout_value": str(stdout.getvalue()),
            }
            self.stdout_queue.put(stream_info)

    def get_tasks(self, collaborator_name: str) -> Tuple:
        """RPC called by a collaborator to determine which tasks to perform.
        Tasks will only be sent to selected collaborators.

        Args:
            collaborator_name (str): Collaborator name which requested tasks.

        Returns:
            next_step (str): Next function to be executed by collaborator
            clone_bytes (bytes): Function execution context for collaborator
        """
        # If requesting collaborator is not registered as connected
        # collaborator, then register it
        if collaborator_name not in self.connected_collaborators:
            logger.info(f"Collaborator {collaborator_name} is connected.")
            self.connected_collaborators.append(collaborator_name)

        logger.debug(
            f"Aggregator GetTasks function reached from collaborator {collaborator_name}..."
        )

        # If queue of requesting collaborator is empty
        while self.__collaborator_tasks_queue[collaborator_name].qsize() == 0:
            # If it is time to then inform the collaborator
            if self.time_to_quit:
                logger.info(f"Sending signal to collaborator {collaborator_name} to shutdown...")
                self.quit_job_sent_to.append(collaborator_name)
                # FIXME: 0, and "" instead of None is just for protobuf compatibility.
                #  Cleaner solution?
                return (
                    0,
                    "",
                    None,
                    Aggregator._get_sleep_time(),
                    self.time_to_quit,
                )

            # If not time to quit then sleep for 10 seconds
            time.sleep(Aggregator._get_sleep_time())

        # Get collaborator step, and clone for requesting collaborator
        next_step, clone = self.__collaborator_tasks_queue[collaborator_name].get()

        self.tasks_sent_to_collaborators += 1
        logger.info(
            "Sending tasks to collaborator"
            + f" {collaborator_name} for round {self.current_round}..."
        )
        return (
            self.current_round,
            next_step,
            dill.dumps(clone),
            0,
            self.time_to_quit,
        )

    def do_task(self, f_name: str) -> Any:
        """Execute aggregator steps until transition.

        Args:
            f_name (str): Aggregator step

        Returns:
            string / None: Next collaborator function or None end of the flow.
        """
        # Set aggregator private attributes to flow object
        self.__set_private_attrs_to_clone(self.flow)

        not_at_transition_point = True
        # Run a loop to execute flow steps until not_at_transition_point
        # is False
        while not_at_transition_point:
            f = getattr(self.flow, f_name)
            # Get the list of parameters of function f
            args = inspect.signature(f)._parameters

            if f.__name__ == "end":
                f()
                # Take the checkpoint of "end" step
                self.__delete_private_attrs_from_clone(
                    self.flow, "Private attributes: Not Available."
                )
                self.call_checkpoint(self.name, self.flow, f)
                self.__set_private_attrs_to_clone(self.flow)
                # Check if all rounds of external loop is executed
                if self.current_round is self.rounds_to_train:
                    # All rounds execute, it is time to quit
                    self.time_to_quit = True
                    # It is time to quit - Break the loop
                    not_at_transition_point = False
                # Start next round of execution
                else:
                    self.current_round += 1
                    logger.info(f"Starting round {self.current_round}...")
                    f_name = "start"
                continue

            selected_clones = ()
            # If function requires arguments then it is join step of the flow
            if len(args) > 0:
                # Check if total number of collaborators and number of
                # selected collaborators are the same
                if len(self.selected_collaborators) != len(self.clones_dict):
                    # Create list of selected collaborator clones
                    selected_clones = ([],)
                    for name, clone in self.clones_dict.items():
                        # Check if collaboraotr is in the list of selected
                        # collaborators
                        if name in self.selected_collaborators:
                            selected_clones[0].append(clone)
                else:
                    # Number of selected collaborators, and number of total
                    # collaborators are same
                    selected_clones = (list(self.clones_dict.values()),)
            # Call the join function with selected collaborators
            # clones are arguments
            f(*selected_clones)

            self.__delete_private_attrs_from_clone(self.flow, "Private attributes: Not Available.")
            # Take the checkpoint of executed step
            self.call_checkpoint(self.name, self.flow, f)
            self.__set_private_attrs_to_clone(self.flow)

            # Next function in the flow
            _, f, parent_func = self.flow.execute_task_args[:3]
            f_name = f.__name__

            self.flow._display_transition_logs(f, parent_func)
            # Transition check
            if aggregator_to_collaborator(f, parent_func):
                # Transition encountered, break the loop
                not_at_transition_point = False

        # Delete aggregator private attribute from flow object
        self.__delete_private_attrs_from_clone(self.flow)

        # Unpack execute_task_args - clones_dict, instance snapshot and kwargs
        self.clones_dict, self.instance_snapshot, self.kwargs = self.flow.execute_task_args[3:]
        if "foreach" in self.kwargs:
            self.flow.filter_exclude_include(f, **self.kwargs)
            self.selected_collaborators = getattr(self.flow, self.kwargs["foreach"])
        else:
            self.kwargs = self.flow.execute_task_args[-1]

        return f_name if f_name != "end" else None

    def send_task_results(
        self,
        collab_name: str,
        round_number: int,
        next_step: str,
        clone_bytes: bytes,
    ) -> None:
        """After collaborator execution, collaborator will call this function
        via gRPc to send next function.

        Args:
            collab_name (str): Collaborator name which is sending results
            round_number (int): Round number for which collaborator is sending
                results
            next_step (str): Next aggregator step in the flow
            clone_bytes (bytes): Collaborator FLSpec object

        Returns:
            None
        """
        # Log a warning if collaborator is sending results for old round
        if round_number is not self.current_round:
            logger.warning(
                f"Collaborator {collab_name} is reporting results"
                f" for the wrong round: {round_number}. Ignoring..."
            )
        else:
            logger.info(f"Collaborator {collab_name} sent task results for round {round_number}.")
        # Unpickle the clone (FLSpec object)
        clone = dill.loads(clone_bytes)
        # Update the clone in clones_dict dictionary
        self.clones_dict[clone.input] = clone
        self.next_step = next_step[0]
        # Sync flow state with clone
        self.flow._foreach_methods = list(set(clone._foreach_methods))
        self.flow.execute_task_args = clone.execute_task_args

        self.collaborators_counter += 1
        # If selected collaborator have sent the results
        if self.collaborators_counter is len(self.selected_collaborators):
            self.collaborators_counter = 0
            # Set the event to inform aggregator to resume the flow execution
            self.collaborator_task_results.set()
            # Empty tasks_sent_to_collaborators list for next time.
            if self.tasks_sent_to_collaborators == len(self.selected_collaborators):
                self.tasks_sent_to_collaborators = 0

    def valid_collaborator_cn_and_id(
        self, cert_common_name: str, collaborator_common_name: str
    ) -> bool:
        """Determine if the collaborator certificate and ID are valid for this
        federation.

        Args:
            cert_common_name: Common name for security certificate
            collaborator_common_name: Common name for collaborator

        Returns:
            bool: True means the collaborator common name matches the name in
                  the security certificate.
        """
        # if self.test_mode_whitelist is None, then the common_name must
        # match collaborator_common_name and be in authorized_cols
        # FIXME: "" instead of None is just for protobuf compatibility.
        #  Cleaner solution?
        if self.single_col_cert_common_name == "":
            return (
                cert_common_name == collaborator_common_name
                and collaborator_common_name in self.authorized_cols
            )
        # otherwise, common_name must be in whitelist and
        # collaborator_common_name must be in authorized_cols
        else:
            return (
                cert_common_name == self.single_col_cert_common_name
                and collaborator_common_name in self.authorized_cols
            )

    def all_quit_jobs_sent(self) -> bool:
        """Assert all quit jobs are sent to collaborators."""
        return set(self.quit_job_sent_to) == set(self.authorized_cols)


the_dragon = """

 ,@@.@@+@@##@,@@@@.`@@#@+  *@@@@ #@##@  `@@#@# @@@@@   @@    @@@@` #@@@ :@@ `@#`@@@#.@
  @@ #@ ,@ +. @@.@* #@ :`   @+*@ .@`+.   @@ *@::@`@@   @@#  @@  #`;@`.@@ @@@`@`#@* +:@`
  @@@@@ ,@@@  @@@@  +@@+    @@@@ .@@@    @@ .@+:@@@:  .;+@` @@ ,;,#@` @@ @@@@@ ,@@@* @
  @@ #@ ,@`*. @@.@@ #@ ,;  `@+,@#.@.*`   @@ ,@::@`@@` @@@@# @@`:@;*@+ @@ @`:@@`@ *@@ `
 .@@`@@,+@+;@.@@ @@`@@;*@  ;@@#@:*@+;@  `@@;@@ #@**@+;@ `@@:`@@@@  @@@@.`@+ .@ +@+@*,@
  `` ``     ` ``  .     `     `      `     `    `  .` `  ``   ``    ``   `       .   `



                                            .**
                                      ;`  `****:
                                     @**`*******
                         ***        +***********;
                        ,@***;` .*:,;************
                        ;***********@@***********
                        ;************************,
                        `*************************
                         *************************
                         ,************************
                          **#*********************
                          *@****`     :**********;
                          +**;          .********.
                          ;*;            `*******#:                       `,:
                                          ****@@@++::                ,,;***.
                                          *@@@**;#;:         +:      **++*,
                                          @***#@@@:          +*;     ,****
                                          @*@+****           ***`     ****,
                                         ,@#******.  ,       ****     **;,**.
                                         * ******** :,       ;*:*+    **  :,**
                                        #  ********::      *,.*:**`   *      ,*;
                                        .  *********:      .+,*:;*:   :      `:**
                                       ;   :********:       ***::**   `       ` **
                                       +   :****::***  ,    *;;::**`             :*
                                      ``   .****::;**:::    *;::::*;              ;*
                                      *     *****::***:.    **::::**               ;:
                                      #     *****;:****     ;*::;***               ,*`
                                      ;     ************`  ,**:****;               ::*
                                      :     *************;:;*;*++:                   *.
                                      :     *****************;*                      `*
                                     `.    `*****************;  :                     *.
                                     .`    .*+************+****;:                     :*
                                     `.    :;+***********+******;`    :              .,*
                                      ;    ::*+*******************. `::              .`:.
                                      +    :::**********************;;:`                *
                                      +    ,::;*************;:::*******.                *
                                      #    `:::+*************:::;********  :,           *
                                      @     :::***************;:;*********;:,           *
                                      @     ::::******:*********************:         ,:*
                                      @     .:::******:;*********************,         :*
                                      #      :::******::******###@*******;;****        *,
                                      #      .::;*****::*****#****@*****;:::***;  ``  **
                                      *       ::;***********+*****+#******::*****,,,,**
                                      :        :;***********#******#******************
                                      .`       `;***********#******+****+************
                                      `,        ***#**@**+***+*****+**************;`
                                       ;         *++**#******#+****+`      `.,..
                                       +         `@***#*******#****#
                                       +          +***@********+**+:
                                       *         .+**+;**;;;**;#**#
                                      ,`         ****@         +*+:
                                      #          +**+         :+**
                                      @         ;**+,       ,***+
                                      #      #@+****      *#****+
                                     `;     @+***+@      `#**+#++
                                     #      #*#@##,      .++:.,#
                                    `*      @#            +.
                                  @@@
                                 # `@
                                  ,                                                        """



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Adagrad optimizer module."""

from typing import Dict, Optional

import numpy as np

from openfl.utilities.optimizers.numpy.base_optimizer import Optimizer


class NumPyAdagrad(Optimizer):
    """Adagrad optimizer implementation.

    Implements the Adagrad optimization algorithm using NumPy. Adagrad is an
    algorithm for gradient-based optimization that adapts the learning rate to
    the parameters, performing smaller updates for parameters associated with
    frequently occurring features, and larger updates for parameters
    associated with infrequent features.

    Original paper: http://jmlr.org/papers/v12/duchi11a.html

    Attributes:
        params (dict, optional): Parameters to be stored for optimization.
        model_interface: Model interface instance to provide parameters.
        learning_rate (float): Tuning parameter that determines the step size
            at each iteration.
        initial_accumulator_value (float): Initial value for squared gradients.
        epsilon (float): Value for computational stability.
    """

    def __init__(
        self,
        *,
        params: Optional[Dict[str, np.ndarray]] = None,
        model_interface=None,
        learning_rate: float = 0.01,
        initial_accumulator_value: float = 0.1,
        epsilon: float = 1e-10,
    ) -> None:
        """Initialize the Adagrad optimizer.

        Args:
            params (dict, optional): Parameters to be stored for optimization.
                Defaults to None.
            model_interface: Model interface instance to provide parameters.
                Defaults to None.
            learning_rate (float, optional): Tuning parameter that determines
                the step size at each iteration. Defaults to 0.01.
            initial_accumulator_value (float, optional): Initial value for
                squared gradients. Defaults to 0.1.
            epsilon (float, optional): Value for computational stability.
                Defaults to 1e-10.

        Raises:
            ValueError: If both params and model_interface are None.
            ValueError: If learning_rate is less than 0.
            ValueError: If initial_accumulator_value is less than 0.
            ValueError: If epsilon is less than or equal to 0.
        """
        super().__init__()

        if model_interface is None and params is None:
            raise ValueError("Should provide one of the params or model_interface")

        if learning_rate < 0:
            raise ValueError(f"Invalid learning rate: {learning_rate}. Learning rate must be >= 0.")
        if initial_accumulator_value < 0:
            raise ValueError(
                f"Invalid initial_accumulator_value value: {initial_accumulator_value}."
                "Initial accumulator value must be >= 0."
            )
        if epsilon <= 0:
            raise ValueError(f"Invalid epsilon value: {epsilon}. Epsilon avalue must be > 0.")

        self.params = params

        if params is None and model_interface is not None:
            self._set_params_from_model(model_interface)

        self.learning_rate = learning_rate
        self.initial_accumulator_value = initial_accumulator_value
        self.epsilon = epsilon

        self.grads_squared = {}
        for param_name in self.params:
            self.grads_squared[param_name] = np.full_like(
                self.params[param_name], self.initial_accumulator_value
            )

    def _update_param(self, grad_name: str, grad: np.ndarray) -> None:
        """
        Update papams by given gradients.

        Args:
            grad_name (str): The name of the gradient.
            grad (np.ndarray): The gradient values.
        """
        self.params[grad_name] -= (
            self.learning_rate * grad / (np.sqrt(self.grads_squared[grad_name]) + self.epsilon)
        )

    def step(self, gradients: Dict[str, np.ndarray]) -> None:
        """Perform a single step for parameter update.

        Implement Adagrad optimizer weights update rule.

        Args:
            gradients (dict): Partial derivatives with respect to optimized
                parameters.

        Raises:
            KeyError: If a key in gradients does not exist in optimized
                parameters.
        """
        for grad_name in gradients:
            if grad_name not in self.grads_squared:
                raise KeyError(f"Key {grad_name} doesn't exist in optimized parameters")

            grad = gradients[grad_name]
            self.grads_squared[grad_name] = self.grads_squared[grad_name] + grad**2
            self._update_param(grad_name, grad)



import logging
import threading
import queue
import grpc
from concurrent.futures import ThreadPoolExecutor
from flwr.proto import grpcadapter_pb2_grpc
from src.grpc.connector.flower.message_conversion import flower_to_openfl_message, openfl_to_flower_message
from multiprocessing import cpu_count
import signal
import psutil
import time

logger = logging.getLogger(__name__)


class FlowerInteropServer(grpcadapter_pb2_grpc.GrpcAdapterServicer):
    """
    FlowerInteropServer is a gRPC server that handles requests from the Flower SuperNode
    and forwards them to the OpenFL Client. It uses a queue-based system to
    ensure that requests are processed sequentially, preventing concurrent
    request handling issues.
    """

    def __init__(self, openfl_client, collaborator_name):
        """
        Initialize.

        Args:
            openfl_client: An instance of the OpenFL Client.
            collaborator_name: The name of the collaborator.
        """
        self.openfl_client = openfl_client
        self.collaborator_name = collaborator_name
        self.end_experiment_callback = None
        self.request_queue = queue.Queue()
        self.processing_thread = threading.Thread(target=self.process_queue)
        self.processing_thread.daemon = True
        self.processing_thread.start()
        self.server = None
        self.termination_event = threading.Event()
        self.port = None

    def set_end_experiment_callback(self, callback):
        self.end_experiment_callback = callback

    def start_server(self, local_server_port):
        """Starts the gRPC server."""
        self.server = grpc.server(ThreadPoolExecutor(max_workers=cpu_count()))
        grpcadapter_pb2_grpc.add_GrpcAdapterServicer_to_server(self, self.server)
        self.port = self.server.add_insecure_port(f'[::]:{local_server_port}')
        self.server.start()
        logger.info(f"OpenFL local gRPC server started, listening on port {self.port}.")

    def get_port(self):
        # Return the port that was assigned
        return self.port

    def stop_server(self):
        """Stops the gRPC server."""
        if self.server:
            logger.info("Shutting down local gRPC server...")
            self.server.stop(0)
            logger.info("local gRPC server stopped.")
            self.termination_event.set()

    def SendReceive(self, request, context):
        """ Handles incoming gRPC requests by putting them into the request queue and waiting for the response.
        Args:
            request: The incoming gRPC request.
            context: The gRPC context.
        Returns:
            The response from the OpenFL server.
        """
        response_queue = queue.Queue()
        self.request_queue.put((request, response_queue))
        return response_queue.get()

    def process_queue(self):
        """
        Continuously processes requests from the request queue. Each request is
        sent to the OpenFL server, and the response is put into the corresponding
        response queue.
        """
        while True:
            request, response_queue = self.request_queue.get()
            openfl_request = flower_to_openfl_message(request)

            # Send request to the OpenFL server
            openfl_response = self.openfl_client.send_message_to_server(openfl_request, self.collaborator_name)

            # Check to end experiment
            if hasattr(openfl_response, 'metadata'):
                if openfl_response.metadata['end_experiment'] == 'True':
                    self.end_experiment_callback()

            # Send response to Flower client
            flower_response = openfl_to_flower_message(openfl_response)
            response_queue.put(flower_response)
            self.request_queue.task_done()

    def handle_signals(self, supernode_process):
        """Sets up signal handlers for graceful shutdown."""
        def signal_handler(_sig, _frame):
            self.terminate_supernode_process(supernode_process)
            self.stop_server()

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

    def terminate_supernode_process(self, supernode_process):
        """Terminates the SuperNode process."""
        if supernode_process.poll() is None:
            try:
                main_subprocess = psutil.Process(supernode_process.pid)
                client_app_processes = main_subprocess.children(recursive=True)

                for client_app_process in client_app_processes:
                    self.terminate_process(client_app_process)

                self.terminate_process(main_subprocess)
                logger.info("SuperNode process terminated.")

            except Exception as e:
                logger.debug(f"Error during graceful shutdown: {e}")
                time.sleep(10)
                supernode_process.kill()
                logger.info("SuperNode process terminated.")
        else:
            logger.info("SuperNode process already terminated.")

    def terminate_process(self, process, timeout=5):
        """Helper function to terminate a process gracefully."""
        try:
            process.terminate()
            process.wait(timeout=timeout)
        except psutil.TimeoutExpired:
            logger.debug(f"Timeout expired while waiting for process {process.pid} to terminate. Killing the process.")
            process.kill()
        except psutil.NoSuchProcess:
            logger.debug(f"Process {process.pid} does not exist. Skipping.")
            pass



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Workspace module."""

import logging
import os
import shutil
import subprocess  # nosec
import sys
import tempfile
from hashlib import sha256
from pathlib import Path
from typing import Union

from click import Choice, echo, group, option, pass_context
from click import Path as ClickPath
from cryptography.hazmat.primitives import serialization

from openfl.cryptography.ca import generate_root_cert, generate_signing_csr, sign_certificate
from openfl.federated.plan import Plan
from openfl.interface import plan
from openfl.interface.cli_helper import CERT_DIR, SITEPACKS, WORKSPACE, print_tree


@group()
@pass_context
def workspace(context):
    """Manage Federated Learning Workspaces."""
    context.obj["group"] = "workspace"


def is_directory_traversal(directory: Union[str, Path]) -> bool:
    """Check for directory traversal.

    Args:
        directory (Union[str, Path]): The directory to check.

    Returns:
        bool: True if directory traversal is detected, False otherwise.
    """
    cwd = os.path.abspath(os.getcwd())
    requested_path = os.path.relpath(directory, start=cwd)
    requested_path = os.path.abspath(requested_path)
    common_prefix = os.path.commonprefix([requested_path, cwd])
    return common_prefix != cwd


def create_dirs(prefix):
    """Create workspace directories.

    Args:
        prefix: The prefix for the directories to be created.
    """

    echo("Creating Workspace Directories")

    (prefix / "cert").mkdir(parents=True, exist_ok=True)  # certifications
    (prefix / "data").mkdir(parents=True, exist_ok=True)  # training data
    (prefix / "logs").mkdir(parents=True, exist_ok=True)  # training logs
    (prefix / "save").mkdir(parents=True, exist_ok=True)  # model weight saves / initialization
    (prefix / "local_state").mkdir(parents=True, exist_ok=True)  # persistent state
    (prefix / "src").mkdir(parents=True, exist_ok=True)  # model code

    shutil.copyfile(WORKSPACE / "workspace" / ".workspace", prefix / ".workspace")


def create_temp(prefix, template):
    """Create workspace templates.

    Args:
        prefix: The prefix for the directories to be created.
        template: The template to use for creating the workspace.
    """

    src = template if os.path.isabs(template) else WORKSPACE / template
    echo(f"Creating Workspace Templates from {src} in {prefix}")
    shutil.copytree(
        src=src,
        dst=prefix,
        dirs_exist_ok=True,
        ignore=shutil.ignore_patterns("__pycache__"),
    )  # from template workspace


def get_templates():
    """Grab the default templates from the distribution.

    Returns:
        list: A list of default templates.
    """

    templates = []
    excluded_dirs = ["workspace", "experimental"]
    for root, _, files in os.walk(WORKSPACE):
        if any(file.endswith(".workspace") for file in files):
            dir_path = os.path.relpath(root, WORKSPACE)
            dir_path = dir_path.replace(os.sep, "/")
            if dir_path and not any(dir_path.startswith(prefix) for prefix in excluded_dirs):
                templates.append(dir_path)
    return templates


@workspace.command(name="create")
@option(
    "--prefix", required=True, help="Path to the new workspace including name.", type=ClickPath()
)
@option(
    "--template",
    required=True,
    type=Choice(get_templates()),
    help="Name of a built-in template to use for creating the workspace.",
)
def create_(prefix, template):
    """Creates a workspace."""
    if is_directory_traversal(prefix):
        echo("Workspace name or path is out of the openfl workspace scope.")
        sys.exit(1)
    create(prefix, template)


def create(prefix, template):
    """Create federated learning workspace.

    Args:
        prefix: The prefix for the directories to be created.
        template: The template to use for creating the workspace.
    """
    prefix = Path(prefix).absolute()

    create_dirs(prefix)
    create_temp(prefix, template)

    apply_template_plan(prefix, template)

    print_tree(prefix, level=3)


@workspace.command(name="import")
@option(
    "--archive",
    required=True,
    help="Path to workspace archive (generated using ``fx workspace export``).",
    type=ClickPath(exists=True),
)
def import_(archive):
    """Imports an FL workspace."""
    dir_path = os.path.basename(os.path.abspath(archive)).split(".")[0]
    shutil.unpack_archive(archive, extract_dir=dir_path)
    echo(f"Imported workspace `{archive}`.")
    echo("You may need to copy your PKI certificates to join the federation.")


@workspace.command(name="certify")
def certify_():
    """Creates a certificate authority (CA) for the federation."""
    certify()


def certify():
    """Create certificate authority for federation."""

    echo("Setting Up Certificate Authority...\n")

    echo("1.  Create Root CA")
    echo("1.1 Create Directories")

    (CERT_DIR / "ca/root-ca/private").mkdir(parents=True, exist_ok=True, mode=0o700)
    (CERT_DIR / "ca/root-ca/db").mkdir(parents=True, exist_ok=True)

    echo("1.2 Create Database")

    with open(CERT_DIR / "ca/root-ca/db/root-ca.db", "w", encoding="utf-8") as f:
        pass  # write empty file
    with open(CERT_DIR / "ca/root-ca/db/root-ca.db.attr", "w", encoding="utf-8") as f:
        pass  # write empty file

    with open(CERT_DIR / "ca/root-ca/db/root-ca.crt.srl", "w", encoding="utf-8") as f:
        f.write("01")  # write file with '01'
    with open(CERT_DIR / "ca/root-ca/db/root-ca.crl.srl", "w", encoding="utf-8") as f:
        f.write("01")  # write file with '01'

    echo("1.3 Create CA Request and Certificate")

    root_crt_path = "ca/root-ca.crt"
    root_key_path = "ca/root-ca/private/root-ca.key"

    root_private_key, root_cert = generate_root_cert()

    # Write root CA certificate to disk
    with open(CERT_DIR / root_crt_path, "wb") as f:
        f.write(
            root_cert.public_bytes(
                encoding=serialization.Encoding.PEM,
            )
        )

    with open(CERT_DIR / root_key_path, "wb") as f:
        f.write(
            root_private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    echo("2.  Create Signing Certificate")
    echo("2.1 Create Directories")

    (CERT_DIR / "ca/signing-ca/private").mkdir(parents=True, exist_ok=True, mode=0o700)
    (CERT_DIR / "ca/signing-ca/db").mkdir(parents=True, exist_ok=True)

    echo("2.2 Create Database")

    with open(CERT_DIR / "ca/signing-ca/db/signing-ca.db", "w", encoding="utf-8") as f:
        pass  # write empty file
    with open(CERT_DIR / "ca/signing-ca/db/signing-ca.db.attr", "w", encoding="utf-8") as f:
        pass  # write empty file

    with open(CERT_DIR / "ca/signing-ca/db/signing-ca.crt.srl", "w", encoding="utf-8") as f:
        f.write("01")  # write file with '01'
    with open(CERT_DIR / "ca/signing-ca/db/signing-ca.crl.srl", "w", encoding="utf-8") as f:
        f.write("01")  # write file with '01'

    echo("2.3 Create Signing Certificate CSR")

    signing_csr_path = "ca/signing-ca.csr"
    signing_crt_path = "ca/signing-ca.crt"
    signing_key_path = "ca/signing-ca/private/signing-ca.key"

    signing_private_key, signing_csr = generate_signing_csr()

    # Write Signing CA CSR to disk
    with open(CERT_DIR / signing_csr_path, "wb") as f:
        f.write(
            signing_csr.public_bytes(
                encoding=serialization.Encoding.PEM,
            )
        )

    with open(CERT_DIR / signing_key_path, "wb") as f:
        f.write(
            signing_private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )

    echo("2.4 Sign Signing Certificate CSR")

    signing_cert = sign_certificate(signing_csr, root_private_key, root_cert.subject, ca=True)

    with open(CERT_DIR / signing_crt_path, "wb") as f:
        f.write(
            signing_cert.public_bytes(
                encoding=serialization.Encoding.PEM,
            )
        )

    echo("3   Create Certificate Chain")

    # create certificate chain file by combining root-ca and signing-ca
    with open(CERT_DIR / "cert_chain.crt", "w", encoding="utf-8") as d:
        with open(CERT_DIR / "ca/root-ca.crt", encoding="utf-8") as s:
            d.write(s.read())
        with open(CERT_DIR / "ca/signing-ca.crt") as s:
            d.write(s.read())

    echo("\nDone.")


def _get_dir_hash(path):
    """Get the hash of a directory.

    Args:
        path (str): The path of the directory.

    Returns:
        str: The hash of the directory.
    """
    hash_ = sha256()
    hash_.update(path.encode("utf-8"))
    hash_ = hash_.hexdigest()
    return hash_


# Commands for workspace packaging and distribution
# -------------------------------------------------


@workspace.command(name="export")
def export_() -> str:
    """
    Exports the OpenFL workspace (in current directory)
    to an archive.

    \b
    The archive contains the following files/directories copied as-is:
        - ``src``: All experiment source code.
        - ``plan``: The FL plan directory.
        - ``save``: Model initial weights.
        - ``requirements.txt``: Package list required for the experiment.

    This command does not copy ``data``, ``logs``, or secrets.

    This command takes no additional arguments.
    """
    plan_file = os.path.abspath(os.path.join("plan", "plan.yaml"))
    if not os.path.isfile(plan_file):
        raise FileNotFoundError(
            f"{plan_file} does not exist in the current directory.\n"
            "Please ensure this command is being run from a workspace."
        )
    plan.freeze_plan(plan_file)

    # Create a staging area.
    workspace_name = os.path.basename(os.getcwd())
    tmp_dir = os.path.join(tempfile.mkdtemp(), "openfl", workspace_name)
    ignore = shutil.ignore_patterns(
        *["__pycache__", "*.crt", "*.key", "*.csr", "*.srl", "*.pem", "*.pbuf"]
    )

    # Export the minimum required files to set up a collaborator
    # os.makedirs(os.path.join(tmp_dir, 'save'), exist_ok=True)
    os.makedirs(os.path.join(tmp_dir, "logs"), exist_ok=True)
    os.makedirs(os.path.join(tmp_dir, "data"), exist_ok=True)
    os.makedirs(os.path.join(tmp_dir, "local_state"), exist_ok=True)
    shutil.copytree("src", os.path.join(tmp_dir, "src"), ignore=ignore)
    shutil.copytree("plan", os.path.join(tmp_dir, "plan"), ignore=ignore)
    shutil.copytree("save", os.path.join(tmp_dir, "save"))
    shutil.copy2("requirements.txt", os.path.join(tmp_dir, "requirements.txt"))

    _ws_identifier_file = ".workspace"
    if not os.path.isfile(_ws_identifier_file):
        openfl_ws_identifier_file = os.path.join(WORKSPACE, "workspace", _ws_identifier_file)
        logging.warning(
            f"`{_ws_identifier_file}` is missing, copying {openfl_ws_identifier_file} as-is."
        )
        shutil.copy2(openfl_ws_identifier_file, tmp_dir)
    shutil.copy2(_ws_identifier_file, tmp_dir)

    # Create Zip archive of directory
    _ARCHIVE_FORMAT = "zip"
    shutil.make_archive(workspace_name, _ARCHIVE_FORMAT, tmp_dir)
    archive = f"{workspace_name}.{_ARCHIVE_FORMAT}"
    logging.info(f"Export: {archive} created")
    return archive


@workspace.command(name="dockerize")
@option(
    "--save",
    is_flag=True,
    default=False,
    help="Export the workspace image as ``<workspace_name>.tar`` file.",
    show_default=True,
)
@option(
    "--rebuild",
    is_flag=True,
    default=False,
    help="If set, rebuilds workspace image with ``--no-cache`` option.",
    show_default=True,
)
@option(
    "--enclave-key",
    "enclave_key",
    type=str,
    required=False,
    help=(
        "Path to an enclave signing key. If not provided, a key will be auto-generated in the "
        "workspace. Note that this command builds a TEE-ready image by default. "
        "You have the flexibility to not run inside a TEE later."
    ),
)
@option(
    "--base-image",
    "base_image",
    required=False,
    default="ghcr.io/securefederatedai/openfl:latest",
    help=(
        "OpenFL base image to use for creating a workspace. "
        "If unspecified, default latest base image will be pulled. "
        "To build a base image from source, follow docker build "
        "`instructions. <https://openfl.readthedocs.io/en/latest/installation.html#using-docker>`_ "
        "Format: ``IMAGE_URI:TAG``"
    ),
    show_default=True,
)
@pass_context
def dockerize_(context, save: bool, rebuild: bool, enclave_key: str, base_image: str):
    """Packages current workspace as a TEE-ready docker image."""

    # Export workspace
    archive = context.invoke(export_)
    workspace_name, _ = archive.split(".")

    # Build workspace image.
    options = []
    options.append("--no-cache" if rebuild else "")
    options.append(f"--build-arg BASE_IMAGE={base_image}")
    options = " ".join(options)
    logging.info(f"Using base image: {base_image}")
    if enclave_key is None:
        _execute(["openssl", "genrsa", "-out", "key.pem", "-3", "3072"])
        enclave_key = os.path.abspath("key.pem")
        logging.info(f"Generated new enclave key: {enclave_key}")
    else:
        enclave_key = os.path.abspath(enclave_key)
        if not os.path.exists(enclave_key):
            raise FileNotFoundError(f"Enclave key `{enclave_key}` does not exist")
        logging.info(f"Using enclave key: {enclave_key}")

    logging.info("Building workspace image")
    ws_image_build_cmd = [
        "docker",
        "build",
        *options.split(),
        "--build-arg",
        f"WORKSPACE_NAME={workspace_name}",
        "--secret",
        f"id=signer-key,src={enclave_key}",
        "-t",
        workspace_name,
        "-f",
        os.path.join(SITEPACKS, "openfl-docker", "Dockerfile.workspace"),
        ".",
    ]
    _execute(ws_image_build_cmd, env_var={"DOCKER_BUILDKIT": "1"})

    # Export workspace as tarball (optional)
    if save:
        logging.info("Saving workspace docker image...")
        save_image_cmd = ["docker", "save", workspace_name, "-o", f"{workspace_name}.tar"]
        _execute(save_image_cmd)
        logging.info(f"Docker image saved to file: {workspace_name}.tar")


def apply_template_plan(prefix, template):
    """Copy plan file from template folder.

    This function unfolds default values from template plan configuration
    and writes the configuration to the current workspace.

    Args:
        prefix: The prefix for the directories to be created.
        template: The template to use for creating the workspace.
    """

    template_plan = Plan.parse(WORKSPACE / template / "plan" / "plan.yaml")

    Plan.dump(prefix / "plan" / "plan.yaml", template_plan.config)


def _execute(cmd: str, env_var: dict = {}, verbose=True) -> None:
    """Executes `cmd` as a subprocess

    Args:
        cmd (str): Command to be executed.

    Raises:
        Exception: If return code is nonzero

    Returns:
        `stdout` of the command as list of messages
    """
    logging.info(f"Executing: {cmd}")
    process = subprocess.Popen(cmd, env=env_var, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    stdout_log = []
    for line in process.stdout:
        msg = line.rstrip().decode("utf-8")
        stdout_log.append(msg)
        if verbose:
            logging.info(msg)

    process.communicate()
    if process.returncode != 0:
        raise Exception(f"`{cmd}` failed with return code {process.returncode}")

    return stdout_log



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


import platform
import shutil
import urllib.request

from click import confirm

VERSION = "0.16.0"

ARCHITECTURE_ALIASES = {
    "x86_64": "amd64",
    "armv6l": "armv6",
    "armv7l": "armv7",
    "aarch64": "arm64",
}

FILE_EXTENSIONS = {"windows": "zip", "linux": "tar.gz"}


def get_system_and_architecture():
    """Get the system and architecture of the machine.

    Returns:
        tuple: The system and architecture of the machine.
    """
    uname_res = platform.uname()
    system = uname_res.system.lower()

    architecture = uname_res.machine.lower()
    architecture = ARCHITECTURE_ALIASES.get(architecture, architecture)

    return system, architecture


def download_step_bin(prefix=".", confirmation=True):
    """
    Download step binaries.

    Args:
        prefix (str, optional): Folder path to download. Defaults to '.'.
        confirmation (bool, optional): Request user confirmation or not.
            Defaults to True.
    """
    system, arch = get_system_and_architecture()
    ext = FILE_EXTENSIONS[system]
    binary = f"step_{system}_{VERSION}_{arch}.{ext}"
    url = f"https://dl.step.sm/gh-release/cli/docs-cli-install/v{VERSION}/{binary}"
    _download(url, prefix, confirmation)


def download_step_ca_bin(prefix=".", confirmation=True):
    """
    Download step-ca binaries.

    Args:
        prefix (str, optional): Folder path to download. Defaults to '.'.
        confirmation (bool, optional): Request user confirmation or not.
            Defaults to True.
    """
    system, arch = get_system_and_architecture()
    ext = FILE_EXTENSIONS[system]
    binary = f"step-ca_{system}_{VERSION}_{arch}.{ext}"
    url = f"https://dl.step.sm/gh-release/certificates/docs-ca-install/v{VERSION}/{binary}"
    _download(url, prefix, confirmation)


def _download(url, prefix, confirmation):
    """Download a file from a URL.

    Args:
        url (str): URL of the file to download.
        prefix (str): Folder path to download.
        confirmation (bool): Request user confirmation or not.
    """
    if confirmation:
        confirm("CA binaries will be downloaded now", default=True, abort=True)
    name = url.split("/")[-1]
    # nosec: private function definition with static urls
    urllib.request.urlretrieve(url, f"{prefix}/{name}")  # nosec
    shutil.unpack_archive(f"{prefix}/{name}", f"{prefix}/step")



"""app-pytorch: A Flower / PyTorch app."""

from flwr.common import Context, ndarrays_to_parameters
from flwr.server import ServerApp, ServerAppComponents, ServerConfig
from flwr.server.strategy import FedAvg
from app_pytorch.task import Net, get_weights

from flwr.server.client_proxy import ClientProxy
from flwr.common import FitRes, EvaluateRes, Scalar, Parameters, parameters_to_ndarrays
from typing import Optional, Union
import numpy as np
from flwr.server.strategy.aggregate import weighted_loss_avg
from flwr.common.logger import log
from logging import WARNING

class SaveModelStrategy(FedAvg):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.largest_loss = 1e9
        self.aggregated_ndarrays = None

    def aggregate_fit(
        self,
        server_round: int,
        results: list[tuple[ClientProxy, FitRes]],
        failures: list[Union[tuple[ClientProxy, FitRes], BaseException]],
    ) -> tuple[Optional[Parameters], dict[str, Scalar]]:
        """Aggregate model weights using weighted average and store checkpoint"""

        # Call aggregate_fit from base class (FedAvg) to aggregate parameters and metrics
        aggregated_parameters, aggregated_metrics = super().aggregate_fit(
            server_round, results, failures
        )

        if aggregated_parameters is not None:
            # Convert `Parameters` to `list[np.ndarray]`
            self.aggregated_ndarrays: list[np.ndarray] = parameters_to_ndarrays(
                aggregated_parameters
            )

            np.savez(f"./save/last.npz", *self.aggregated_ndarrays)

        return aggregated_parameters, aggregated_metrics

    def aggregate_evaluate(
        self,
        server_round: int,
        results: list[tuple[ClientProxy, EvaluateRes]],
        failures: list[Union[tuple[ClientProxy, EvaluateRes], BaseException]],
    ) -> tuple[Optional[float], dict[str, Scalar]]:
        """Aggregate evaluation losses using weighted average."""
        if not results:
            return None, {}
        # Do not aggregate if there are failures and failures are not accepted
        if not self.accept_failures and failures:
            return None, {}

        # Aggregate loss
        loss_aggregated = weighted_loss_avg(
            [
                (evaluate_res.num_examples, evaluate_res.loss)
                for _, evaluate_res in results
            ]
        )


        # Aggregate custom metrics if aggregation fn was provided
        metrics_aggregated = {}
        if self.evaluate_metrics_aggregation_fn:
            eval_metrics = [(res.num_examples, res.metrics) for _, res in results]
            metrics_aggregated = self.evaluate_metrics_aggregation_fn(eval_metrics)
        elif server_round == 1:  # Only log this warning once
            log(WARNING, "No evaluate_metrics_aggregation_fn provided")

        if loss_aggregated < self.largest_loss:
            self.largest_loss = loss_aggregated
            np.savez(f"./save/best.npz", *self.aggregated_ndarrays)

        return loss_aggregated, metrics_aggregated


def server_fn(context: Context):
    # Read from config
    num_rounds = context.run_config["num-server-rounds"]
    fraction_fit = context.run_config["fraction-fit"]

    # Initialize model parameters
    ndarrays = get_weights(Net())
    parameters = ndarrays_to_parameters(ndarrays)

    # Define strategy
    strategy = SaveModelStrategy(
        fraction_fit=fraction_fit,
        fraction_evaluate=1.0,
        min_available_clients=2,
        initial_parameters=parameters,
    )
    config = ServerConfig(num_rounds=num_rounds)

    return ServerAppComponents(strategy=strategy, config=config)


# Create ServerApp
app = ServerApp(server_fn=server_fn)



# Copyright (C) 2020-2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from keras.models import model_from_json
from keras_contrib.layers import InstanceNormalization

from openfl.federated import KerasTaskRunner
from keras import backend as K
import gdown
import hashlib

MODEL_JSON_HASH = "c35cfa990000ad87825f182460395ec5d1437a707bd33de4df55d45664e94214"
MODEL_WEIGHTS_HASH = "cd5e52d42e2c6d737e370fb0e673aec5d257134e127c0e59478f11676fa327a5"


def dice_coefficient(y_true, y_pred, smooth=1.0):
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    return (2.0 * intersection + smooth) / (K.sum(y_true_f) + K.sum(y_pred_f) + smooth)


def dice_coefficient_loss(y_true, y_pred):
    return -dice_coefficient(y_true, y_pred)


def compute_file_hash(file_path):
    with open(file_path, "rb") as file:
        data = file.read()
        file_hash = hashlib.sha256(data).hexdigest()
    return file_hash


class KerasHippmapp3r(KerasTaskRunner):
    """A basic convolutional neural network model."""

    def __init__(self, **kwargs):
        """
        Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function
        """
        super().__init__(**kwargs)

        weights_id = "1_VEOScLGyr1qV-t-zggq8Lxwgf_z-IpQ"
        gdown.download(id=weights_id, output="model.h5")
        json_id = "1RUE3Cw_rpKnKfwlu75kLbkcr9hde9nV4"
        gdown.download(id=json_id, output="model.json")

        model_json_hash = compute_file_hash("model.json")
        if model_json_hash != MODEL_JSON_HASH:
            raise ValueError("Model JSON file hash does not match expected value.")

        model_weights_hash = compute_file_hash("model.h5")
        if model_weights_hash != MODEL_WEIGHTS_HASH:
            raise ValueError("Model weights file hash does not match expected value.")

        self.model = self.build_model(model_json="model.json", model_weights="model.h5", **kwargs)

        self.initialize_tensorkeys_for_functions()

    def build_model(self, model_json, model_weights, **kwargs):
        """
        Define the model architecture.

        Args:
            model_json (str): Path to model json config
            model_weights (str): Path to model weights

        Returns:
            keras.models.Sequential: The model defined in Keras

        """
        custom_objects = {}
        custom_objects["InstanceNormalization"] = InstanceNormalization
        json_file = open(model_json, "r")
        loaded_model_json = json_file.read()
        json_file.close()
        model = model_from_json(loaded_model_json, custom_objects=custom_objects)
        model.load_weights(model_weights)

        model.compile(loss=dice_coefficient_loss, optimizer="adam", metrics=[dice_coefficient])

        return model



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Collaborator module."""

import logging
from src.grpc import connector

from openfl.component import Collaborator

logger = logging.getLogger(__name__)

class CollaboratorFlower(Collaborator):
    r"""The Collaborator object class.

    Attributes:
        collaborator_name (str): The common name for the collaborator.
        aggregator_uuid (str): The unique id for the client.
        federation_uuid (str): The unique id for the federation.
        client (object): The client object.
        task_runner (object): The task runner object.
        task_config (dict): The task configuration.
        opt_treatment (str)*: The optimizer state treatment.
        device_assignment_policy (str): [Deprecated] The device assignment policy.
        use_delta_updates (bool)*: If True, only model delta gets sent. If False,
            whole model gets sent to collaborator.
        compression_pipeline (object): The compression pipeline.
        db_store_rounds (int): The number of rounds to store in the database.
        single_col_cert_common_name (str): The common name for the single
            column certificate.

    .. note::
        \* - Plan setting.
    """

    def do_task(self, task, round_number) -> dict:
        """Perform the specified task.

        Args:
            task (list_of_str): List of tasks.
            round_number (int): Actual round number.

        Returns:
            A dictionary of reportable metrics of the current collaborator for the task.
        """
        # map this task to an actual function name and kwargs
        if isinstance(task, str):
            task_name = task
        else:
            task_name = task.name
        func_name = self.task_config[task_name]["function"]
        kwargs = self.task_config[task_name]["kwargs"]
        if func_name=="start_client_adapter":
            if hasattr(self.task_runner, func_name):
                method = getattr(self.task_runner, func_name)
                if callable(method):
                    framework = self.task_config['settings']["connect_to"]
                    LocalGRPCServer = connector.get_interop_server(framework)
                    interop_server = LocalGRPCServer(self.client, self.collaborator_name)
                    method(interop_server, **kwargs)
                    self.client.send_local_task_results(round_number, task_name)
                    metrics = {f'{self.collaborator_name}/start_client_adapter': 'Completed'}
                    return metrics
                else:
                    raise AttributeError(f"{func_name} is not callable on {self.task_runner}")
            else:
                raise AttributeError(f"{func_name} does not exist on {self.task_runner}")



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Proto utils."""

import logging

from openfl.protocols import base_pb2
from openfl.utilities import TensorKey

logger = logging.getLogger(__name__)


def model_proto_to_bytes_and_metadata(model_proto):
    """Convert the model protobuf to bytes and metadata.

    Args:
        model_proto: The protobuf of the model.

    Returns:
        bytes_dict: A dictionary where the keys are tensor names and the
            values are the corresponding tensor data in bytes.
        metadata_dict: A dictionary where the keys are tensor names and the
            values are the corresponding tensor metadata.
        round_number: The round number for the model.
    """
    bytes_dict = {}
    metadata_dict = {}
    round_number = None
    for tensor_proto in model_proto.tensors:
        bytes_dict[tensor_proto.name] = tensor_proto.data_bytes
        metadata_dict[tensor_proto.name] = [
            {
                "int_to_float": proto.int_to_float,
                "int_list": proto.int_list,
                "bool_list": proto.bool_list,
            }
            for proto in tensor_proto.transformer_metadata
        ]
        if round_number is None:
            round_number = tensor_proto.round_number
        else:
            assert round_number == tensor_proto.round_number, (
                f"Round numbers in model are inconsistent: {round_number} "
                f"and {tensor_proto.round_number}"
            )
    return bytes_dict, metadata_dict, round_number


def bytes_and_metadata_to_model_proto(bytes_dict, model_id, model_version, is_delta, metadata_dict):
    """
    Convert bytes and metadata to model protobuf.

    Args:
        bytes_dict: A dictionary where the keys are tensor names and the
            values are the corresponding tensor data in bytes.
        model_id: The ID of the model.
        model_version: The version of the model.
        is_delta: A flag indicating whether the model is a delta model.
        metadata_dict: A dictionary where the keys are tensor names and the
            values are the corresponding tensor metadata.

    Returns:
        model_proto: The protobuf of the model.
    """
    model_header = ModelHeader(id=model_id, version=model_version, is_delta=is_delta)  # noqa: F821

    tensor_protos = []
    for key, data_bytes in bytes_dict.items():
        transformer_metadata = metadata_dict[key]
        metadata_protos = []
        for metadata in transformer_metadata:
            if metadata.get("int_to_float") is not None:
                int_to_float = metadata.get("int_to_float")
            else:
                int_to_float = {}

            if metadata.get("int_list") is not None:
                int_list = metadata.get("int_list")
            else:
                int_list = []

            if metadata.get("bool_list") is not None:
                bool_list = metadata.get("bool_list")
            else:
                bool_list = []
            metadata_protos.append(
                base_pb2.MetadataProto(
                    int_to_float=int_to_float,
                    int_list=int_list,
                    bool_list=bool_list,
                )
            )
        tensor_protos.append(
            TensorProto(  # noqa: F821
                name=key,
                data_bytes=data_bytes,
                transformer_metadata=metadata_protos,
            )
        )
    return base_pb2.ModelProto(header=model_header, tensors=tensor_protos)


def construct_named_tensor(tensor_key, nparray, transformer_metadata, lossless):
    """Construct named tensor.

    Args:
        tensor_key: The key of the tensor.
        nparray: The numpy array representing the tensor data.
        transformer_metadata: The transformer metadata for the tensor.
        lossless: A flag indicating whether the tensor is lossless.

    Returns:
        named_tensor: The named tensor.
    """
    metadata_protos = []
    for metadata in transformer_metadata:
        if metadata.get("int_to_float") is not None:
            int_to_float = metadata.get("int_to_float")
        else:
            int_to_float = {}

        if metadata.get("int_list") is not None:
            int_list = metadata.get("int_list")
        else:
            int_list = []

        if metadata.get("bool_list") is not None:
            bool_list = metadata.get("bool_list")
        else:
            bool_list = []
        metadata_protos.append(
            base_pb2.MetadataProto(
                int_to_float=int_to_float,
                int_list=int_list,
                bool_list=bool_list,
            )
        )

    tensor_name, origin, round_number, report, tags = tensor_key

    return base_pb2.NamedTensor(
        name=tensor_name,
        round_number=round_number,
        lossless=lossless,
        report=report,
        tags=tags,
        transformer_metadata=metadata_protos,
        data_bytes=nparray,
    )


def construct_proto(tensor_dict, model_id, model_version, is_delta, compression_pipeline):
    """Construct proto.

    Args:
        tensor_dict: A dictionary where the keys are tensor names and the
            values are the corresponding tensors.
        model_id: The ID of the model.
        model_version: The version of the model.
        is_delta: A flag indicating whether the model is a delta model.
        compression_pipeline: The compression pipeline for the model.

    Returns:
        model_proto: The protobuf of the model.
    """
    # compress the arrays in the tensor_dict, and form the model proto
    # TODO: Hold-out tensors from the compression pipeline.
    bytes_dict = {}
    metadata_dict = {}
    for key, array in tensor_dict.items():
        bytes_dict[key], metadata_dict[key] = compression_pipeline.forward(data=array)

    # convert the compressed_tensor_dict and metadata to protobuf, and make the new model proto
    model_proto = bytes_and_metadata_to_model_proto(
        bytes_dict=bytes_dict,
        model_id=model_id,
        model_version=model_version,
        is_delta=is_delta,
        metadata_dict=metadata_dict,
    )
    return model_proto


def construct_model_proto(tensor_dict, round_number, tensor_pipe):
    """Construct model proto from tensor dict.

    Args:
        tensor_dict: A dictionary where the keys are tensor names and the
            values are the corresponding tensors.
        round_number: The round number for the model.
        tensor_pipe: The tensor pipe for the model.

    Returns:
        model_proto: The protobuf of the model.
    """
    # compress the arrays in the tensor_dict, and form the model proto
    # TODO: Hold-out tensors from the tensor compression pipeline.
    named_tensors = []
    for key, nparray in tensor_dict.items():
        bytes_data, transformer_metadata = tensor_pipe.forward(data=nparray)
        tensor_key = TensorKey(key, "agg", round_number, False, ("model",))
        named_tensors.append(
            construct_named_tensor(
                tensor_key,
                bytes_data,
                transformer_metadata,
                lossless=True,
            )
        )

    return base_pb2.ModelProto(tensors=named_tensors)


def deconstruct_model_proto(model_proto, compression_pipeline):
    """Deconstruct model proto.

    This function takes a model protobuf and a compression pipeline,
    and deconstructs the protobuf into a dictionary of tensors and a round
    number.

    Args:
        model_proto: The protobuf of the model.
        compression_pipeline: The compression pipeline for the model.

    Returns:
        tensor_dict: A dictionary where the keys are tensor names and the
        values are the corresponding tensors.
        round_number: The round number for the model.
    """
    # extract the tensor_dict and metadata
    bytes_dict, metadata_dict, round_number = model_proto_to_bytes_and_metadata(model_proto)

    # decompress the tensors
    # TODO: Handle tensors meant to be held-out from the compression pipeline
    #  (currently none are held out).
    tensor_dict = {}
    for key in bytes_dict:
        tensor_dict[key] = compression_pipeline.backward(
            data=bytes_dict[key], transformer_metadata=metadata_dict[key]
        )
    return tensor_dict, round_number


def deconstruct_proto(model_proto, compression_pipeline):
    """Deconstruct the protobuf.

    This function takes a model protobuf and a compression pipeline, and
    deconstructs the protobuf into a dictionary of tensors.

    Args:
        model_proto: The protobuf of the model.
        compression_pipeline: The compression pipeline for the model.

    Returns:
        tensor_dict: A dictionary where the keys are tensor names and the
            values are the corresponding tensors.
    """
    # extract the tensor_dict and metadata
    bytes_dict, metadata_dict = model_proto_to_bytes_and_metadata(model_proto)

    # decompress the tensors
    # TODO: Handle tensors meant to be held-out from the compression pipeline
    #  (currently none are held out).
    tensor_dict = {}
    for key in bytes_dict:
        tensor_dict[key] = compression_pipeline.backward(
            data=bytes_dict[key], transformer_metadata=metadata_dict[key]
        )
    return tensor_dict


def load_proto(fpath):
    """Load the protobuf.

    Args:
        fpath: The file path of the protobuf.

    Returns:
        model: The protobuf of the model.
    """
    with open(fpath, "rb") as f:
        loaded = f.read()
        model = base_pb2.ModelProto().FromString(loaded)
        return model


def dump_proto(model_proto, fpath):
    """Dump the protobuf to a file.

    Args:
        model_proto: The protobuf of the model.
        fpath: The file path to dump the protobuf.
    """
    s = model_proto.SerializeToString()
    with open(fpath, "wb") as f:
        f.write(s)


def datastream_to_proto(proto, stream):
    """Convert the datastream to the protobuf.

    Args:
        proto: The protobuf to be filled with the data stream.
        stream: The data stream.
        logger (optional): The logger for logging information.

    Returns:
        proto: The protobuf filled with the data stream.
    """
    npbytes = bytearray()
    for chunk in stream:
        npbytes.extend(chunk.npbytes)

    if len(npbytes) > 0:
        proto.ParseFromString(bytes(npbytes))
        return proto
    else:
        raise RuntimeError(f"Received empty stream message of type {type(proto)}")


def proto_to_datastream(proto, max_buffer_size=(2 * 1024 * 1024)):
    """Convert the protobuf to the datastream for the remote connection.

    Args:
        proto: The protobuf to be converted into a data stream.
        logger: The logger for logging information.
        max_buffer_size (optional): The maximum buffer size for the data
            stream. Defaults to 2*1024*1024.

    Yields:
        reply: Chunks of the data stream for the remote connection.
    """
    npbytes = proto.SerializeToString()
    data_size = len(npbytes)
    buffer_size = data_size if max_buffer_size > data_size else max_buffer_size

    for i in range(0, data_size, buffer_size):
        chunk = npbytes[i : i + buffer_size]
        reply = base_pb2.DataStream(npbytes=chunk, size=len(chunk))
        yield reply


def get_headers(context) -> dict:
    """Get headers from context.

    Args:
        context: The context containing the headers.

    Returns:
        headers: A dictionary where the keys are header names and the
            values are the corresponding header values.
    """
    return {header[0]: header[1] for header in context.invocation_metadata()}



# Copyright (C) 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Director CLI."""

import logging
import sys
from pathlib import Path

import click
from click import Path as ClickPath
from click import group, option, pass_context
from dynaconf import Validator

from openfl.experimental.workflow.component.director import Director
from openfl.experimental.workflow.transport import DirectorGRPCServer
from openfl.utilities import merge_configs
from openfl.utilities.path_check import is_directory_traversal

logger = logging.getLogger(__name__)


@group()
@pass_context
def director(context):
    """Manage Federated Learning Director.

    Args:
        context (click.core.Context): Click context.
    """
    context.obj["group"] = "director"


@director.command(name="start")
@option(
    "-c",
    "--director-config-path",
    default="director.yaml",
    help="The director config file path",
    type=ClickPath(exists=True),
)
@option(
    "--tls/--disable-tls",
    default=True,
    is_flag=True,
    help="Use TLS or not (By default TLS is enabled)",
)
@option(
    "-rc",
    "--root-cert-path",
    "root_certificate",
    required=False,
    type=ClickPath(exists=True),
    default=None,
    help="Path to a root CA cert",
)
@option(
    "-pk",
    "--private-key-path",
    "private_key",
    required=False,
    type=ClickPath(exists=True),
    default=None,
    help="Path to a private key",
)
@option(
    "-oc",
    "--public-cert-path",
    "certificate",
    required=False,
    type=ClickPath(exists=True),
    default=None,
    help="Path to a signed certificate",
)
def start(director_config_path, tls, root_certificate, private_key, certificate):
    """Start the director service.

    Args:
        director_config_path (str): The director config file path.
        tls (bool): Use TLS or not.
        root_certificate (str): Path to a root CA cert.
        private_key (str): Path to a private key.
        certificate (str): Path to a signed certificate.
    """

    director_config_path = Path(director_config_path).absolute()
    logger.info("🧿 Starting the Director Service.")
    if is_directory_traversal(director_config_path):
        click.echo("The director config file path is out of the openfl workspace scope.")
        sys.exit(1)
    config = merge_configs(
        settings_files=director_config_path,
        overwrite_dict={
            "root_certificate": root_certificate,
            "private_key": private_key,
            "certificate": certificate,
        },
        validators=[
            Validator("settings.listen_host", default="localhost"),
            Validator("settings.listen_port", default=50051, gte=1024, lte=65535),
            Validator("settings.install_requirements", default=False),
            Validator(
                "settings.envoy_health_check_period",
                default=60,  # in seconds
                gte=1,
                lte=24 * 60 * 60,
            ),
        ],
    )

    if config.root_certificate:
        config.root_certificate = Path(config.root_certificate).absolute()

    if config.private_key:
        config.private_key = Path(config.private_key).absolute()

    if config.certificate:
        config.certificate = Path(config.certificate).absolute()

    director_server = DirectorGRPCServer(
        director_cls=Director,
        tls=tls,
        root_certificate=config.root_certificate,
        private_key=config.private_key,
        certificate=config.certificate,
        listen_host=config.settings.listen_host,
        listen_port=config.settings.listen_port,
        envoy_health_check_period=config.settings.envoy_health_check_period,
        install_requirements=config.settings.install_requirements,
        director_config=director_config_path,
    )
    director_server.start()



# Copyright 2025 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from abc import abstractmethod
from typing import Any, Dict

import torch


class LabelMapper:
    """
    A utility class for mapping class names (labels) to unique integer indices and vice versa.
    This ensures consistent labeling across data sources and processes by maintaining a
    bidirectional mapping between labels and their corresponding indices.
    """

    def __init__(self):
        self.label_to_idx = {}
        self.idx_to_label = {}

    def get_label_index(self, label: str) -> int:
        """Assigns or retrieves the index of a label."""
        if label not in self.label_to_idx:
            new_index = len(self.label_to_idx)
            self.label_to_idx[label] = new_index
            self.idx_to_label[new_index] = label
        return self.label_to_idx[label]

    def get_label_name(self, index: int) -> str:
        """Retrieves the original label name from an index."""
        return self.idx_to_label.get(index, None)


class FolderDataset(torch.utils.data.Dataset):
    def __init__(self, label_mapper: LabelMapper, transform=None):
        """
        Args:
            label_mapper (LabelMapper): LabelMapper object to map class names to indices.
            transform (callable, optional): Transformations to apply to images.
        """
        self.transform = transform
        self.label_mapper = label_mapper

        # Build the dataset
        self.samples = self._load_samples()

    def _get_label(self, file_path):
        """Get the label for a given file path."""
        label_name = file_path.split("/")[-2] if len(file_path.split("/")) > 1 else None
        return self.label_mapper.get_label_index(label_name)

    def _load_samples(self):
        """Loads all file paths and their inferred labels"""
        return [
            (file_path, self._get_label(file_path))
            for file_path in self.datasource.enumerate_files()
        ]

    @abstractmethod
    def load_file(self, file_path):
        """Load a file from the dataset."""
        pass

    def __getitem__(self, index) -> Dict[str, Any]:
        file_path, label = self.samples[index]
        file_data = self.load_file(str(file_path))

        if self.transform:
            file_data = self.transform(file_data)

        return {"data": file_data, "label": label, "path": str(file_path)}

    def __len__(self):
        return len(self.samples)



import os
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

import src.patch.patch_flwr_build

import re
from flwr.cli.app import app
if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
    sys.exit(app())



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""PyTorchDataLoader module."""

from openfl.federated.data.loader import DataLoader


class GaNDLFDataLoaderWrapper(DataLoader):
    """A class used to represent a data loader for the Generally Nuanced Deep
    Learning Framework (GaNDLF).

    Attributes:
        train_csv (str): Path to the training CSV file.
        val_csv (str): Path to the validation CSV file.
        train_dataloader (DataLoader): DataLoader object for the training data.
        val_dataloader (DataLoader): DataLoader object for the validation data.
        feature_shape (tuple): Shape of an example feature array.
    """

    def __init__(self, data_path, feature_shape):
        """Initializes the GaNDLFDataLoaderWrapper object.

        Args:
            data_path (str): The path to the directory containing the data.
            feature_shape (tuple): The shape of an example feature array.
        """
        if "inference" in data_path:
            self.train_csv = None
        else:
            self.train_csv = data_path + "/train.csv"
        self.val_csv = data_path + "/valid.csv"
        self.train_dataloader = None
        self.val_dataloader = None
        self.feature_shape = feature_shape

    def set_dataloaders(self, train_dataloader, val_dataloader):
        """Sets the data loaders for the training and validation data.

        Args:
            train_dataloader (DataLoader): The DataLoader object for the
                training data.
            val_dataloader (DataLoader): The DataLoader object for the
                validation data.
        """
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader

    def get_feature_shape(self):
        """Returns the shape of an example feature array.

        Returns:
            tuple: The shape of an example feature array.
        """
        return self.feature_shape

    def get_train_loader(self, batch_size=None, num_batches=None):
        """Returns the data loader for the training data.

        Args:
            batch_size (int, optional): The batch size for the data loader
                (default is None).
            num_batches (int, optional): The number of batches for the data
                loader (default is None).

        Returns:
            DataLoader: The DataLoader object for the training data.
        """
        return self.train_dataloader

    def get_valid_loader(self, batch_size=None):
        """Returns the data loader for the validation data.

        Args:
            batch_size (int, optional): The batch size for the data loader
                (default is None).

        Returns:
            DataLoader: The DataLoader object for the validation data.
        """
        return self.val_dataloader

    def get_train_data_size(self):
        """Returns the total number of training samples.

        Returns:
            int: The total number of training samples.
        """
        return len(self.train_dataloader.dataset)

    def get_valid_data_size(self):
        """Returns the total number of validation samples.

        Returns:
            int: The total number of validation samples.
        """
        return len(self.val_dataloader.dataset)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Base abstract optimizer class module."""

import abc
from importlib import import_module
from os.path import splitext
from typing import Dict

from numpy import ndarray

from openfl.plugins.frameworks_adapters.framework_adapter_interface import (
    FrameworkAdapterPluginInterface,
)


class Optimizer(abc.ABC):
    """Base abstract optimizer class.

    This class serves as a base class for all optimizers. It defines the basic
    structure that all derived optimizer classes should follow.
    It includes an abstract method `step` that must be implemented by any
    concrete optimizer class.
    """

    @abc.abstractmethod
    def step(self, gradients: Dict[str, ndarray]) -> None:
        """Perform a single step for parameter update.

        This method should be overridden by all subclasses to implement the
        specific optimization algorithm.

        Args:
            gradients (dict): Partial derivatives with respect to optimized
                parameters.
        """
        pass

    def _set_params_from_model(self, model_interface):
        """
        Eject and store model parameters.

        This method is used to extract the parameters from the provided model
        interface and store them in the optimizer.

        Args:
            model_interface: The model interface instance to provide
                parameters.
        """
        class_name = splitext(model_interface.framework_plugin)[1].strip(".")
        module_path = splitext(model_interface.framework_plugin)[0]
        framework_adapter = import_module(module_path)
        framework_adapter_plugin: FrameworkAdapterPluginInterface = getattr(
            framework_adapter, class_name, None
        )
        self.params: Dict[str, ndarray] = framework_adapter_plugin.get_tensor_dict(
            model_interface.provide_model()
        )



import flwr.cli.build
from flwr.cli.build import write_to_zip, get_fab_filename
from typing import Annotated, Optional
import typer
from pathlib import Path
from flwr.cli.utils import is_valid_project_name
from flwr.cli.config_utils import load_and_validate
import tempfile
import zipfile
from flwr.common.constant import FAB_ALLOWED_EXTENSIONS
import shutil
import tomli_w
import hashlib
import os
from src.util import is_safe_path

def build(
    app: Annotated[
        Optional[Path],
        typer.Option(help="Path of the Flower App to bundle into a FAB"),
    ] = None,
) -> tuple[str, str]:
    """Build a Flower App into a Flower App Bundle (FAB).

    You can run ``flwr build`` without any arguments to bundle the app located in the
    current directory. Alternatively, you can specify a path using the ``--app``
    option to bundle an app located at the provided path. For example:

    ``flwr build --app ./apps/flower-hello-world``.
    """
    ### PATCH ###
    # # REASONING: original code writes to /tmp/ by default. Writing to flwr_home allows us to consolidate written files
    # # This is useful for running in an SGX enclave with Gramine since we need to strictly control allowed/trusted files
    flwr_home = os.getenv("FLWR_HOME")
    #################################

    if app is None:
        app = Path.cwd()

    app = app.resolve()
    if not app.is_dir():
        typer.secho(
            f"❌ The path {app} is not a valid path to a Flower app.",
            fg=typer.colors.RED,
            bold=True,
        )
        raise typer.Exit(code=1)

    if not is_valid_project_name(app.name):
        typer.secho(
            f"❌ The project name {app.name} is invalid, "
            "a valid project name must start with a letter, "
            "and can only contain letters, digits, and hyphens.",
            fg=typer.colors.RED,
            bold=True,
        )
        raise typer.Exit(code=1)

    conf, errors, warnings = load_and_validate(app / "pyproject.toml")
    if conf is None:
        typer.secho(
            "Project configuration could not be loaded.\npyproject.toml is invalid:\n"
            + "\n".join([f"- {line}" for line in errors]),
            fg=typer.colors.RED,
            bold=True,
        )
        raise typer.Exit(code=1)

    if warnings:
        typer.secho(
            "Project configuration is missing the following "
            "recommended properties:\n" + "\n".join([f"- {line}" for line in warnings]),
            fg=typer.colors.RED,
            bold=True,
        )

    # Load .gitignore rules if present
    ignore_spec = flwr.cli.build._load_gitignore(app)

    list_file_content = ""

    # Remove the 'federations' field from 'tool.flwr' if it exists
    if (
        "tool" in conf
        and "flwr" in conf["tool"]
        and "federations" in conf["tool"]["flwr"]
    ):
        del conf["tool"]["flwr"]["federations"]

    toml_contents = tomli_w.dumps(conf)

    with tempfile.NamedTemporaryFile(suffix=".zip", delete=False) as temp_file:
        temp_filename = temp_file.name

        with zipfile.ZipFile(temp_filename, "w", zipfile.ZIP_DEFLATED) as fab_file:
            write_to_zip(fab_file, "pyproject.toml", toml_contents)

            # Continue with adding other files
            all_files = [
                f
                for f in app.rglob("*")
                if not ignore_spec.match_file(f)
                and f.name != temp_filename
                and f.suffix in FAB_ALLOWED_EXTENSIONS
                and f.name != "pyproject.toml"  # Exclude the original pyproject.toml
            ]

            all_files.sort()

            for file_path in all_files:
                # Read the file content manually
                with open(file_path, "rb") as f:
                    file_contents = f.read()

                archive_path = file_path.relative_to(app)
                write_to_zip(fab_file, str(archive_path), file_contents)

                # Calculate file info
                sha256_hash = hashlib.sha256(file_contents).hexdigest()
                file_size_bits = os.path.getsize(file_path) * 8  # size in bits
                list_file_content += f"{archive_path},{sha256_hash},{file_size_bits}\n"

            # Add CONTENT and CONTENT.jwt to the zip file
            write_to_zip(fab_file, ".info/CONTENT", list_file_content)

    # Get hash of FAB file
    content = Path(temp_filename).read_bytes()
    fab_hash = hashlib.sha256(content).hexdigest()

    # Set the name of the zip file
    fab_filename = get_fab_filename(conf, fab_hash)

    ### PATCH ###
    # # REASONING: original code writes to /tmp/ by default. Writing to flwr_home allows us to consolidate written files
    if not os.path.isdir(flwr_home):
        raise ValueError("Invalid directory")

    if not is_safe_path(fab_filename):
        raise ValueError("Invalid filename")

    final_path = os.path.join(flwr_home, fab_filename)
    shutil.move(temp_filename, final_path)
    #################################

    typer.secho(
        f"🎊 Successfully built {fab_filename}", fg=typer.colors.GREEN, bold=True
    )

    ### PATCH ###
    # return final_path
    return final_path, fab_hash
    ################

flwr.cli.build.build = build



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""Adam adaptive aggregation module."""

from typing import Dict, Optional, Tuple

import numpy as np

from openfl.interface.aggregation_functions.core import AdaptiveAggregation, AggregationFunction
from openfl.interface.aggregation_functions.weighted_average import WeightedAverage
from openfl.utilities.optimizers.numpy import NumPyAdam

DEFAULT_AGG_FUNC = WeightedAverage()


class AdamAdaptiveAggregation(AdaptiveAggregation):
    """Adam adaptive Federated Aggregation funtcion."""

    def __init__(
        self,
        *,
        agg_func: AggregationFunction = DEFAULT_AGG_FUNC,
        params: Optional[Dict[str, np.ndarray]] = None,
        model_interface=None,
        learning_rate: float = 0.01,
        betas: Tuple[float, float] = (0.9, 0.999),
        initial_accumulator_value: float = 0.0,
        epsilon: float = 1e-8,
    ) -> None:
        """Initialize the AdamAdaptiveAggregation object.

        Args:
            agg_func (AggregationFunction): Aggregate function for aggregating
                parameters that are not inside the optimizer (default:
                WeightedAverage()).
            params (Optional[Dict[str, np.ndarray]]): Parameters to be stored
                for optimization.
            model_interface: Model interface instance to provide parameters.
            learning_rate (float): Tuning parameter that determines
                the step size at each iteration.
            betas (Tuple[float, float]): Coefficients used for computing
                running averages of gradient and its square.
            initial_accumulator_value (float): Initial value for gradients
                and squared gradients.
            epsilon (float): Value for computational stability.
        """
        opt = NumPyAdam(
            params=params,
            model_interface=model_interface,
            learning_rate=learning_rate,
            betas=betas,
            initial_accumulator_value=initial_accumulator_value,
            epsilon=epsilon,
        )
        super().__init__(opt, agg_func)



# Copyright 2020-2024 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


"""XGBoostTaskRunner module."""

import json
import logging

import numpy as np
import xgboost as xgb
from sklearn.metrics import accuracy_score

from openfl.federated.task.runner import TaskRunner
from openfl.utilities import Metric, TensorKey, change_tags
from openfl.utilities.split import split_tensor_dict_for_holdouts

logger = logging.getLogger(__name__)


def check_precision_loss(converted_data, original_data):
    """
    Checks for precision loss during conversion to float32 and back.

    Parameters:
    converted_data (np.ndarray): The data that has been converted to float32.
    original_data (list): The original data to be checked for precision loss.
    """
    # Convert the float32 array back to bytes and decode to JSON
    reconstructed_bytes = converted_data.astype(np.uint8).tobytes()
    reconstructed_json = reconstructed_bytes.decode("utf-8")
    reconstructed_data = json.loads(reconstructed_json)

    assert type(original_data) is type(reconstructed_data), (
        "Reconstructed datatype does not match original."
    )

    # Compare the original and reconstructed data
    if original_data != reconstructed_data:
        logger.warning("Precision loss detected during conversion.")


class XGBoostTaskRunner(TaskRunner):
    def __init__(self, **kwargs):
        """
        A class to manage XGBoost tasks in a federated learning environment.

        This class inherits from TaskRunner and provides methods to initialize and manage
        the global model and required tensor keys for XGBoost tasks.

        Attributes:
            bst (xgb.Booster): gradient-boosted tree model for XGBoost.
            global_model (np.ndarray): The numpy array containing global XGBoost tree.
            required_tensorkeys_for_function (dict): A dictionary to store required tensor keys
                for each function.
        """
        super().__init__(**kwargs)
        self.bst = None
        self.global_model = None
        self.required_tensorkeys_for_function = {}

    def rebuild_model(self, input_tensor_dict):
        """
        Rebuilds the model using the provided input tensor dictionary.

        This method checks if the 'local_tree' key in the input tensor dictionary is either a
        non-empty numpy array. If this condition is met, it updates the internal tensor dictionary
        with the provided input.

        Parameters:
            input_tensor_dict (dict): A dictionary containing tensor data.
            It must include the key 'local_tree'

        Returns:
        None
        """
        if (
            isinstance(input_tensor_dict["local_tree"], np.ndarray)
            and input_tensor_dict["local_tree"].size != 0
        ):
            self.set_tensor_dict(input_tensor_dict)

    def validate_task(self, col_name, round_num, input_tensor_dict, **kwargs):
        """Validate Task.

        Run validation of the model on the local data.

        Args:
            col_name (str): Name of the collaborator.
            round_num (int): What round is it.
            input_tensor_dict (dict): Required input tensors (for model).
            **kwargs: Additional parameters.

        Returns:
            global_output_dict (dict):  Tensors to send back to the aggregator.
            local_output_dict (dict):   Tensors to maintain in the local TensorDB.
        """
        data = self.data_loader.get_valid_dmatrix()

        # during agg validation, rebuild self.bst from global model. during local validation,
        # continue with trained self.bst, no need to rebuild
        if kwargs["apply"] != "local":
            self.rebuild_model(input_tensor_dict)

        # if self.bst is still None after rebuilding, then there was no initial global model, so
        # set metric to 0
        if self.bst is None:
            # for first round agg validation, there is no model so set metric to 0
            # TODO: this is not robust, especially if using a loss metric
            metric = Metric(name="accuracy", value=np.array(0))
        else:
            metric = self.validate_(data)

        origin = col_name
        suffix = "validate"
        if kwargs["apply"] == "local":
            suffix += "_local"
        else:
            suffix += "_agg"
        tags = ("metric",)
        tags = change_tags(tags, add_field=suffix)

        #  validate function
        output_tensor_dict = {TensorKey(metric.name, origin, round_num, True, tags): metric.value}

        # Empty list represents metrics that should only be stored locally
        return output_tensor_dict, {}

    def train_task(
        self,
        col_name,
        round_num,
        input_tensor_dict,
        **kwargs,
    ):
        """Train batches task.

        Train the model on the requested number of batches.

        Args:
            col_name (str): Name of the collaborator.
            round_num (int): What round is it.
            input_tensor_dict (dict): Required input tensors (for model).
            **kwargs: Additional parameters.

        Returns:
            global_output_dict (dict):  Tensors to send back to the aggregator.
            local_output_dict (dict):   Tensors to maintain in the local
                TensorDB.
        """
        self.rebuild_model(input_tensor_dict)
        data = self.data_loader.get_train_dmatrix()
        metric = self.train_(data)
        # Output metric tensors (scalar)
        origin = col_name
        tags = ("trained",)
        output_metric_dict = {
            TensorKey(metric.name, origin, round_num, True, ("metric",)): metric.value
        }

        # output model tensors (Doesn't include TensorKey)
        output_model_dict = self.get_tensor_dict()
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            output_model_dict, **self.tensor_dict_split_fn_kwargs
        )

        # Create global tensorkeys
        global_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags): nparray
            for tensor_name, nparray in global_model_dict.items()
        }
        # Create tensorkeys that should stay local
        local_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num, False, tags): nparray
            for tensor_name, nparray in local_model_dict.items()
        }
        # The train/validate aggregated function of the next round will look
        # for the updated model parameters.
        # This ensures they will be resolved locally
        next_local_tensorkey_model_dict = {
            TensorKey(tensor_name, origin, round_num + 1, False, ("model",)): nparray
            for tensor_name, nparray in local_model_dict.items()
        }

        global_tensor_dict = {
            **output_metric_dict,
            **global_tensorkey_model_dict,
        }
        local_tensor_dict = {
            **local_tensorkey_model_dict,
            **next_local_tensorkey_model_dict,
        }

        return global_tensor_dict, local_tensor_dict

    def get_tensor_dict(self, with_opt_vars=False):
        """
        Retrieves the tensor dictionary containing the model's tree structure.

        This method returns a dictionary with the key 'local_tree', which contains the model's tree
        structure as a numpy array. If the model has not been initialized (`self.bst` is None), it
        returns an empty numpy array. If the global model is not set or is empty, it returns the
        entire model as a numpy array. Otherwise, it returns only the trees added in the latest
        training session.

        Parameters:
        with_opt_vars (bool): N/A for XGBoost (Default=False).

        Returns:
            dict: A dictionary with the key 'local_tree' containing the model's tree structure as a
            numpy array.
        """

        if self.bst is None:
            # For initializing tensor dict
            return {"local_tree": np.array([], dtype=np.float32)}

        booster_array = self.bst.save_raw("json")
        booster_dict = json.loads(booster_array)

        if (
            isinstance(self.global_model, np.ndarray) and self.global_model.size == 0
        ) or self.global_model is None:
            booster_float32_array = np.frombuffer(booster_array, dtype=np.uint8).astype(np.float32)
            return {"local_tree": booster_float32_array}

        global_model_byte_array = bytearray(self.global_model.astype(np.uint8).tobytes())
        global_model_booster_dict = json.loads(global_model_byte_array)
        num_global_trees = int(
            global_model_booster_dict["learner"]["gradient_booster"]["model"]["gbtree_model_param"][
                "num_trees"
            ]
        )
        num_total_trees = int(
            booster_dict["learner"]["gradient_booster"]["model"]["gbtree_model_param"]["num_trees"]
        )

        # Calculate the number of trees added in the latest training
        num_latest_trees = num_total_trees - num_global_trees
        latest_trees = booster_dict["learner"]["gradient_booster"]["model"]["trees"][
            -num_latest_trees:
        ]

        latest_trees_json = json.dumps(latest_trees)
        latest_trees_bytes = latest_trees_json.encode("utf-8")
        latest_trees_float32_array = np.frombuffer(latest_trees_bytes, dtype=np.uint8).astype(
            np.float32
        )

        check_precision_loss(latest_trees_float32_array, original_data=latest_trees)

        return {"local_tree": latest_trees_float32_array}

    def get_required_tensorkeys_for_function(self, func_name, **kwargs):
        """Get the required tensors for specified function that could be called
        as part of a task. By default, this is just all of the layers and
        optimizer of the model.

        Args:
            func_name (str): The function name.

        Returns:
            list : [TensorKey].
        """
        if func_name == "validate_task":
            local_model = "apply=" + str(kwargs["apply"])
            return self.required_tensorkeys_for_function[func_name][local_model]
        else:
            return self.required_tensorkeys_for_function[func_name]

    def initialize_tensorkeys_for_functions(self, with_opt_vars=False):
        """Set the required tensors for all publicly accessible task methods.

        By default, this is just all of the layers and optimizer of the model.
        Custom tensors should be added to this function.

        Args:
            with_opt_vars (bool): with_opt_vars (bool): N/A for XGBoost (Default=False).

        Returns:
            None
        """
        output_model_dict = self.get_tensor_dict()
        global_model_dict, local_model_dict = split_tensor_dict_for_holdouts(
            output_model_dict, **self.tensor_dict_split_fn_kwargs
        )
        global_model_dict_val = global_model_dict
        local_model_dict_val = local_model_dict

        self.required_tensorkeys_for_function["train_task"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict
        ]
        self.required_tensorkeys_for_function["train_task"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict
        ]

        self.required_tensorkeys_for_function["train_task"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict
        ]
        self.required_tensorkeys_for_function["train_task"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict
        ]

        # Validation may be performed on local or aggregated (global) model,
        # so there is an extra lookup dimension for kwargs
        self.required_tensorkeys_for_function["validate_task"] = {}
        # TODO This is not stateless. The optimizer will not be
        self.required_tensorkeys_for_function["validate_task"]["apply=local"] = [
            TensorKey(tensor_name, "LOCAL", 0, False, ("trained",))
            for tensor_name in {**global_model_dict_val, **local_model_dict_val}
        ]
        self.required_tensorkeys_for_function["validate_task"]["apply=global"] = [
            TensorKey(tensor_name, "GLOBAL", 0, False, ("model",))
            for tensor_name in global_model_dict_val
        ]
        self.required_tensorkeys_for_function["validate_task"]["apply=global"] += [
            TensorKey(tensor_name, "LOCAL", 0, False, ("model",))
            for tensor_name in local_model_dict_val
        ]

    def set_tensor_dict(self, tensor_dict, with_opt_vars=False):
        """Set the tensor dictionary.

        Args:
            tensor_dict (dict): The tensor dictionary.
            with_opt_vars (bool): N/A for XGBoost (Default=False).
        """
        # The with_opt_vars argument is not used in this method
        self.global_model = tensor_dict["local_tree"]
        if (
            isinstance(self.global_model, np.ndarray) and self.global_model.size == 0
        ) or self.global_model is None:
            raise ValueError("The model does not exist or is empty.")
        else:
            global_model_byte_array = bytearray(self.global_model.astype(np.uint8).tobytes())
            self.bst = xgb.Booster()
            self.bst.load_model(global_model_byte_array)

    def load_native(
        self,
        filepath,
        **kwargs,
    ):
        """
        Load XGBooster from a file specified by filepath.

        Args:
            filepath (str): Path to XGB booster file.
            **kwargs: Additional keyword arguments.
        """
        self.bst = xgb.Booster()
        self.bst.load_model(filepath)

    def save_native(
        self,
        filepath,
        **kwargs,
    ):
        """Save XGB booster to file.

        Args:
            filepath (str): Path to pickle file to be created by booster.save_model().
            **kwargs: Additional parameters.

        Returns:
            None
        """
        self.bst.save_model(filepath)

    def train_(self, data) -> Metric:
        """
        Train the XGBoost model.

        Args:
            train_dataloader (dict): A dictionary containing the training data with keys 'dmatrix'.

        Returns:
            Metric: A Metric object containing the training loss.
        """
        dtrain = data["dmatrix"]
        evals = [(dtrain, "train")]
        evals_result = {}

        self.bst = xgb.train(
            self.params,
            dtrain,
            self.num_rounds,
            xgb_model=self.bst,
            evals=evals,
            evals_result=evals_result,
            verbose_eval=False,
        )

        loss = evals_result["train"]["logloss"][-1]
        return Metric(name=self.loss_fn.__name__, value=np.array(loss))

    def validate_(self, data) -> Metric:
        """
        Validate the XGBoost model.

        Args:
            validation_dataloader (dict): A dictionary containing the validation data with keys
            'dmatrix' and 'labels'.

        Returns:
            Metric: A Metric object containing the validation accuracy.
        """
        dtest = data["dmatrix"]
        y_test = data["labels"]
        preds = self.bst.predict(dtest)
        y_pred_binary = np.where(preds > 0.5, 1, 0)
        acc = accuracy_score(y_test, y_pred_binary)

        return Metric(name="accuracy", value=np.array(acc))



# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
@author: Shay Vargaftik (VMware Research),
shayv@vmware.com; vargaftik@gmail.com
@author: Yaniv Ben-Itzhak (VMware Research),
ybenitzhak@vmware.com; yaniv.benizhak@gmail.com


EdenPipeline module.


EDEN is an unbiased lossy compression method that uses a random rotation
    followed by deterministic quantization and scaling.
EDEN provides strong theoretical guarantees, as described in the following
    ICML 2022 paper:

"EDEN: Communication-Efficient and Robust Distributed Mean Estimation for
    Federated Learning"
Shay Vargaftik, Ran Ben Basat, Amit Portnoy, Gal Mendelson, Yaniv Ben Itzhak,
Michael Mitzenmacher,
Proceedings of the 39th International Conference on Machine Learning,
PMLR 162:21984-22014, 2022.

https://proceedings.mlr.press/v162/vargaftik22a.html

------------------------------------------------------------------------------------------------------------------------

In order to use and configure EDEN use the following lines in plan.yaml:

compression_pipeline :
  defaults : plan/defaults/compression_pipeline.yaml
  template : openfl.pipelines.EdenPipeline
  settings :
    n_bits : <number of bits per coordinate>
    device: <cpu|cuda:0|cuda:1|...>
    dim_threshold: 1000 #EDEN compresses layers that their dimension is above
        the dim_threshold, use 1000 as default
"""

import copy as co

import numpy as np
import torch

from openfl.pipelines.pipeline import Float32NumpyArrayToBytes, TransformationPipeline, Transformer


class Eden:
    """Eden class for quantization.

    This class is responsible for quantizing tensors using the Eden method.

    Attributes:
        device (str): The device to be used for quantization ('cpu' or 'cuda').
        centroids (dict): A dictionary mapping the number of bits to the
            corresponding centroids.
        boundaries (dict): A dictionary mapping the number of bits to the
            corresponding boundaries.
        nbits (int): The number of bits per coordinate for quantization.
        num_hadamard (int): The number of Hadamard transforms to employ.
        max_padding_overhead (float): The maximum overhead that is allowed for
            padding the vector.
    """

    def __init__(self, nbits=8, device="cpu"):
        """Initialize Eden.

        Args:
            nbits (int, optional): The number of bits per coordinate for
                quantization. Defaults to 8.
            device (str, optional): The device to be used for quantization
                ('cpu' or 'cuda'). Defaults to 'cpu'.
        """

        def gen_normal_centroids_and_boundaries(device):
            """Generates the centroids and boundaries for the quantization
            process.

            This function generates the centroids and boundaries used in the
            quantization process based on the specified device. The centroids
            are generated for different numbers of bits, and the boundaries
            are calculated based on these centroids.

            Args:
                device (str): The device to be used for quantization
                    ('cpu' or 'cuda').

            Returns:
                tuple: A tuple containing two dictionaries. The first
                    dictionary maps the number of bits to the corresponding
                    centroids, and the second dictionary maps the number of
                    bits to the corresponding boundaries.
            """

            # half-normal lloyd-max centroids
            centroids = {}
            centroids[1] = [0.7978845608028654]
            centroids[2] = [0.4527800398860679, 1.5104176087114887]
            centroids[3] = [
                0.24509416307340598,
                0.7560052489539643,
                1.3439092613750225,
                2.151945669890335,
            ]
            centroids[4] = [
                0.12839501671105813,
                0.38804823445328507,
                0.6567589957631145,
                0.9423402689122875,
                1.2562309480263467,
                1.6180460517130526,
                2.069016730231837,
                2.732588804065177,
            ]
            centroids[5] = [
                0.06588962234909321,
                0.1980516892038791,
                0.3313780514298761,
                0.4666991751197207,
                0.6049331689395434,
                0.7471351317890572,
                0.89456439585444,
                1.0487823813655852,
                1.2118032120324,
                1.3863389353626248,
                1.576226389073775,
                1.7872312118858462,
                2.0287259913633036,
                2.3177364021261493,
                2.69111557955431,
                3.260726295605043,
            ]
            centroids[6] = [
                0.0334094558802581,
                0.1002781217139195,
                0.16729660990171974,
                0.23456656976873475,
                0.3021922894403614,
                0.37028193328115516,
                0.4389488009177737,
                0.5083127587538033,
                0.5785018460645791,
                0.6496542452315348,
                0.7219204720694183,
                0.7954660529025513,
                0.870474868055092,
                0.9471530930156288,
                1.0257343133937524,
                1.1064859596918581,
                1.1897175711327463,
                1.2757916223519965,
                1.3651378971823598,
                1.458272959944728,
                1.5558274659528346,
                1.6585847114298427,
                1.7675371481292605,
                1.8839718992293555,
                2.009604894545278,
                2.146803022259123,
                2.2989727412973995,
                2.471294740528467,
                2.6722617014102585,
                2.91739146530985,
                3.2404166403241677,
                3.7440690236964755,
            ]
            centroids[7] = [
                0.016828143177728235,
                0.05049075396896167,
                0.08417241989671888,
                0.11788596825032507,
                0.1516442630131618,
                0.18546025708680833,
                0.21934708340331643,
                0.25331807190834565,
                0.2873868062260947,
                0.32156710392315796,
                0.355873075050329,
                0.39031926330596733,
                0.4249205523979007,
                0.4596922300454219,
                0.49465018161031576,
                0.5298108436256188,
                0.565191195643323,
                0.600808970989236,
                0.6366826613981411,
                0.6728315674936343,
                0.7092759460939766,
                0.746037126679468,
                0.7831375375631398,
                0.8206007832455021,
                0.858451939611374,
                0.896717615963322,
                0.9354260757626341,
                0.9746074842160436,
                1.0142940678300427,
                1.054520418037026,
                1.0953237719213182,
                1.1367442623434032,
                1.1788252655205043,
                1.2216138763870124,
                1.26516137869917,
                1.309523700469555,
                1.3547621051156036,
                1.4009441065262136,
                1.448144252238147,
                1.4964451375010575,
                1.5459387008934842,
                1.596727786313424,
                1.6489283062238074,
                1.7026711624156725,
                1.7581051606756466,
                1.8154009933798645,
                1.8747553268072956,
                1.9363967204122827,
                2.0005932433837565,
                2.0676621538384503,
                2.1379832427349696,
                2.212016460501213,
                2.2903268704925304,
                2.3736203164211713,
                2.4627959084523208,
                2.5590234991374485,
                2.663867022558051,
                2.7794919110540777,
                2.909021527386642,
                3.0572161028423737,
                3.231896182843021,
                3.4473810105937095,
                3.7348571053691555,
                4.1895219330235225,
            ]
            centroids[8] = [
                0.008445974137017219,
                0.025338726226901278,
                0.042233889994651476,
                0.05913307399220878,
                0.07603788791797023,
                0.09294994306815242,
                0.10987089037069565,
                0.12680234584461386,
                0.1437459285205906,
                0.16070326074968388,
                0.1776760066764216,
                0.19466583496246115,
                0.21167441946986007,
                0.22870343946322488,
                0.24575458029044564,
                0.2628295721769575,
                0.2799301528634766,
                0.29705806782573063,
                0.3142150709211129,
                0.3314029639954903,
                0.34862355883476864,
                0.3658786774238477,
                0.3831701926964899,
                0.40049998943716425,
                0.4178699650069057,
                0.4352820704086704,
                0.45273827097956804,
                0.4702405882876,
                0.48779106011037887,
                0.505391740756901,
                0.5230447441905988,
                0.5407522460590347,
                0.558516486141511,
                0.5763396823538222,
                0.5942241184949506,
                0.6121721459546814,
                0.6301861414640443,
                0.6482685527755422,
                0.6664219019236218,
                0.684648787627676,
                0.7029517931200633,
                0.7213336286470308,
                0.7397970881081071,
                0.7583450032075904,
                0.7769802937007926,
                0.7957059197645721,
                0.8145249861674053,
                0.8334407494351099,
                0.8524564651728141,
                0.8715754936480047,
                0.8908013031010308,
                0.9101374749919184,
                0.9295877653215154,
                0.9491559977740125,
                0.9688461234581733,
                0.9886622867721733,
                1.0086087121824747,
                1.028689768268861,
                1.0489101021225093,
                1.0692743940997251,
                1.0897875553561465,
                1.1104547388972044,
                1.1312812154370708,
                1.1522725891384287,
                1.173434599389649,
                1.1947731980672593,
                1.2162947131430126,
                1.238005717146854,
                1.2599130381874064,
                1.2820237696510286,
                1.304345369166531,
                1.3268857708606756,
                1.349653145284911,
                1.3726560932224416,
                1.3959037693197867,
                1.419405726021264,
                1.4431719292973744,
                1.4672129964566984,
                1.4915401336751468,
                1.5161650628244996,
                1.541100284490976,
                1.5663591473033147,
                1.5919556551358922,
                1.6179046397057497,
                1.6442219553485078,
                1.6709244249695359,
                1.6980300628044107,
                1.7255580190748743,
                1.7535288357430767,
                1.7819645728459763,
                1.81088895442524,
                1.8403273195729115,
                1.870306964218662,
                1.9008577747790962,
                1.9320118435829472,
                1.9638039107009146,
                1.9962716117712092,
                2.0294560760505993,
                2.0634026367482017,
                2.0981611002741527,
                2.133785932225919,
                2.170336784741086,
                2.2078803102947337,
                2.2464908293749546,
                2.286250990303635,
                2.327254033532845,
                2.369604977942217,
                2.4134218838650208,
                2.458840003415269,
                2.506014300608167,
                2.5551242195294983,
                2.6063787537827645,
                2.660023038604595,
                2.716347847697055,
                2.7757011083910723,
                2.838504606698991,
                2.9052776685316117,
                2.976670770545963,
                3.0535115393558603,
                3.136880130166507,
                3.2282236667414654,
                3.3295406612081644,
                3.443713971315384,
                3.5751595986789093,
                3.7311414987004117,
                3.9249650523739246,
                4.185630113705256,
                4.601871059539151,
            ]

            # normal centroids
            for i in centroids:
                centroids[i] = torch.Tensor([-j for j in centroids[i][::-1]] + centroids[i]).to(
                    device
                )

            # centroids to bin boundaries
            def gen_boundaries(centroids):
                return [(a + b) / 2 for a, b in zip(centroids[:-1], centroids[1:])]

            # bin boundaries
            boundaries = {}
            for i in centroids:
                boundaries[i] = torch.Tensor(gen_boundaries(centroids[i])).to(device)

            return centroids, boundaries

        # device - 'cpu' or 'cuda'
        self.device = device

        # generate metadata for compression
        self.centroids, self.boundaries = gen_normal_centroids_and_boundaries(self.device)

        # bits per coordinate
        if nbits not in [1, 2, 3, 4, 5, 6, 7, 8]:
            raise Exception("nbits value is not supported")
        self.nbits = int(nbits)

        # number of hadamard transforms to employ
        self.num_hadamard = 2

        # The maximum overhead that is allowed for padding the vector
        self.max_padding_overhead = 0.1

        # TODO: add as a configurable parameter to YAML?
        # 2 is default and always works,
        # but 1 can be used for non-adversarial inputs and thus run faster

    def rand_diag(self, size, seed):
        """Generate a random diagonal matrix.

        Args:
            size (int): The size of the matrix.
            seed (int): The seed for the random number generator.

        Returns:
            A random diagonal matrix.
        """
        bools_in_float32 = 8

        shift = 32 // bools_in_float32
        threshold = 1 << (shift - 1)
        mask = (1 << shift) - 1

        size_scaled = size // bools_in_float32 + (size % bools_in_float32 != 0)
        mask32 = (1 << 32) - 1

        # hash seed and then limit its size to prevent overflow
        seed = (seed * 1664525 + 1013904223) & mask32
        seed = (seed * 8121 + 28411) & mask32

        r = torch.arange(end=size_scaled, device=self.device) + seed

        # LCG (https://en.wikipedia.org/wiki/Linear_congruential_generator)
        r = (1103515245 * r + 12345 + seed) & mask32
        r = (1140671485 * r + 12820163 + seed) & mask32

        # SplitMix (https://dl.acm.org/doi/10.1145/2714064.2660195)
        r += 0x9E3779B9
        r = (r ^ (r >> 16)) * 0x85EBCA6B & mask32
        r = (r ^ (r >> 13)) * 0xC2B2AE35 & mask32
        r = (r ^ (r >> 16)) & mask32

        res = torch.zeros(size_scaled * bools_in_float32, device=self.device)

        s = 0
        for _ in range(bools_in_float32):
            res[s : s + size_scaled] = r & mask
            s += size_scaled
            r >>= shift

        # convert to signs
        res = 2 * (res >= threshold) - 1

        return res[:size]

    def hadamard(self, vec):
        """Apply the Hadamard transform to a vector.

        Args:
            vec: The vector to be transformed.

        Returns:
            The transformed vector.
        """
        d = vec.numel()
        if d & (d - 1) != 0:
            raise Exception("input numel must be a power of 2")

        h = 2
        while h <= d:
            hf = h // 2
            vec = vec.view(d // h, h)
            vec[:, :hf] = vec[:, :hf] + vec[:, hf : 2 * hf]
            vec[:, hf : 2 * hf] = vec[:, :hf] - 2 * vec[:, hf : 2 * hf]
            h *= 2
        vec /= np.sqrt(d)

        return vec.view(-1)

    def rht(self, vec, seed):
        """Apply the randomized Hadamard transform to a vector.

        Args:
            vec: The vector to be transformed.
            seed (int): The seed for the random number generator.

        Returns:
            The transformed vector.
        """
        vec = vec * self.rand_diag(size=vec.numel(), seed=seed)
        vec = self.hadamard(vec)

        return vec

    def irht(self, vec, seed):
        """Apply the inverse randomized Hadamard transform to a vector.

        Args:
            vec: The vector to be transformed.
            seed (int): The seed for the random number generator.

        Returns:
            The transformed vector.
        """
        vec = self.hadamard(vec)
        vec = vec * self.rand_diag(size=vec.numel(), seed=seed)

        return vec

    def quantize(self, vec):
        """Quantize a vector.

        Args:
            vec: The vector to be quantized.

        Returns:
            bins: The quantized values of the vector.
            scale: The scale factor for the quantization.
        """
        vec_norm = torch.norm(vec, 2)

        if vec_norm > 0:
            normalized = vec * (vec.numel() ** 0.5) / vec_norm
            bins = torch.bucketize(normalized, self.boundaries[self.nbits])
            scale = vec_norm**2 / torch.dot(torch.take(self.centroids[self.nbits], bins), vec)

            if not torch.isnan(scale):
                return bins, scale

        return torch.zeros(vec.numel(), device=self.device), torch.tensor([0])

    def compress_slice(self, vec, seed):
        """Compress a slice of a vector.

        Args:
            vec: The slice of the vector to be compressed.
            seed (int): The seed for the random number generator.

        Returns:
            bins: The compressed values of the slice.
            scale: The scale factor for the compression.
            dim: The dimension of the slice.
        """
        dim = vec.numel()

        if not dim & (dim - 1) == 0 or dim < 8:
            padded_dim = max(int(2 ** (np.ceil(np.log2(dim)))), 8)
            padded_vec = torch.zeros(padded_dim, device=self.device)
            padded_vec[:dim] = vec

            vec = padded_vec

        for i in range(self.num_hadamard):
            vec = self.rht(vec, seed + i)

        bins, scale = self.quantize(vec)

        return bins, float(scale.cpu().numpy()), vec.numel()

    def compress(self, vec, seed):
        """Compress a vector.

        Args:
            vec: The vector to be compressed.
            seed (int): The seed for the random number generator.

        Returns:
            int_array: The compressed values of the vector.
            scale_list: The list of scale factors for the compression.
            dim_list: The list of dimensions for the compression.
            total_dim: The total dimension of the vector.
        """

        def low_po2(n):
            if not n:
                return 0
            return 2 ** int(np.log2(n))

        def high_po2(n):
            if not n:
                return 0
            return 2 ** (int(np.ceil(np.log2(n))))

        vec = torch.Tensor(vec.flatten()).to(self.device)

        res_bins = []
        res_scale = []
        res_dim = []

        dim = remaining = vec.numel()
        curr_index = 0

        while (high_po2(remaining) - remaining) / dim > self.max_padding_overhead:
            low = low_po2(remaining)

            slice_bins, slice_scale, slice_dim = self.compress_slice(
                vec[curr_index : curr_index + low], seed
            )

            res_bins.append(slice_bins)
            res_scale.append(slice_scale)
            res_dim.append(slice_dim)

            curr_index += low
            remaining -= low

        slice_bins, slice_scale, slice_dim = self.compress_slice(vec[curr_index:], seed)

        res_bins.append(slice_bins)
        res_scale.append(slice_scale)
        res_dim.append(slice_dim)

        all_bins = torch.cat(res_bins)
        all_bins = self.to_bits(all_bins)

        return all_bins.cpu().numpy(), res_scale, res_dim, vec.numel()

    def decompress_slice(self, bins, scale, dim, seed):
        """Decompress a slice of a vector.

        Args:
            bins: The compressed values of the slice.
            scale: The scale factor for the decompression.
            dim: The dimension of the slice.
            seed (int): The seed for the random number generator.

        Returns:
            The decompressed slice of the vector.
        """
        vec = torch.take(self.centroids[self.nbits], bins)

        for i in range(self.num_hadamard):
            vec = self.irht(vec, int(seed + (self.num_hadamard - 1) - i))

        return (scale * vec)[:dim]

    def decompress(self, bins, metadata):
        """Decompress a vector.

        Args:
            bins: The compressed values of the vector.
            metadata: The metadata for the decompression.

        Returns:
            The decompressed vector.
        """
        bins = self.from_bits(torch.Tensor(bins).to(self.device)).long().flatten()

        seed = int(metadata[0])
        total_dim = int(metadata[1])

        curr_index = 0
        vec = []

        for k in range(2, max(metadata.keys()) + 1, 2):
            scale = metadata[k]
            dim = int(metadata[k + 1])
            vec.append(self.decompress_slice(bins[curr_index : curr_index + dim], scale, dim, seed))
            curr_index += dim

        vec = torch.cat(vec)
        vec = vec[:total_dim]

        return vec.cpu().numpy()

    def to_bits(self, int_bool_vec):
        """Convert a vector of integers to bits.

        Packing the quantization values to bytes.

        Args:
            int_bool_vec: The vector of integers to be converted.

        Returns:
            The bit vector.
        """

        def to_bits_h(ibv):
            n = ibv.numel()
            ibv = ibv.view(n // 8, 8).int()

            bv = torch.zeros(n // 8, dtype=torch.uint8, device=self.device)
            for i in range(8):
                bv += ibv[:, i] * (2**i)

            return bv

        l_unit = int_bool_vec.numel() // 8
        bit_vec = torch.zeros(l_unit * self.nbits, dtype=torch.uint8)

        for i in range(self.nbits):
            bit_vec[l_unit * i : l_unit * (i + 1)] = to_bits_h((int_bool_vec % 2 != 0).int())
            int_bool_vec = torch.div(int_bool_vec, 2, rounding_mode="floor")

        return bit_vec

    def from_bits(self, bit_vec):
        """Convert a bit vector to integers.

        Unpacking bytes to quantization values.

        Args:
            bit_vec: The bit vector to be converted.

        Returns:
            The vector of integers.
        """

        def from_bits_h(bv):
            n = bv.numel()

            iv = torch.zeros((8, n)).to(self.device)
            for i in range(8):
                temp = bv.clone()
                iv[i] = (torch.div(temp, 2**i, rounding_mode="floor") % 2 != 0).int()

            return iv.T.reshape(-1)

        bit_vec = bit_vec.view(self.nbits, bit_vec.numel() // self.nbits).to(self.device)
        int_vecs = torch.zeros(bit_vec.numel() // self.nbits * 8).to(self.device)

        for i in range(self.nbits):
            int_vecs += 2**i * from_bits_h(bit_vec[i])

        return int_vecs.reshape(1, -1)


class EdenTransformer(Transformer):
    """Eden transformer class for quantizing input data.

    This class is a transformer that uses the Eden method for quantization.

    Attributes:
        n_bits (int): The number of bits per coordinate for quantization.
        dim_threshold (int): The threshold for the dimension of the data. Data
            with dimensions less than this threshold are not compressed.
        device (str): The device to be used for quantization ('cpu' or 'cuda').
        eden (Eden): The Eden object for quantization.
        no_comp (Float32NumpyArrayToBytes): The transformer for data that are
            not compressed.
    """

    def __init__(self, n_bits=8, dim_threshold=100, device="cpu"):
        """Initialize EdenTransformer.

        Args:
            n_bits (int, optional): The number of bits per coordinate for
                quantization. Defaults to 8.
            dim_threshold (int, optional): The threshold for the dimension of
                the data. Data with dimensions less than this threshold are
                not compressed. Defaults to 100.
            device (str, optional): The device to be used for quantization
                ('cpu' or 'cuda'). Defaults to 'cpu'.
        """
        self.lossy = True
        self.eden = Eden(nbits=n_bits, device=device)

        print(
            f"*** Using EdenTransformer with params: {n_bits} bits,"
            f"dim_threshold: {dim_threshold}, {device} device ***"
        )

        self.dim_threshold = dim_threshold
        self.no_comp = Float32NumpyArrayToBytes()

    def forward(self, data, **kwargs):
        """Quantize data.

        Args:
            data: The data to be quantized.

        Returns:
            The quantized data and the metadata for the quantization.
        """
        # TODO: can be simplified if have access to a unique feature of the participant (e.g., ID)
        seed = (hash(sum(data.flatten()) * 13 + 7) + np.random.randint(1, 2**16)) % (2**16)
        seed = int(float(seed))
        metadata = {"int_list": list(data.shape)}

        if data.size > self.dim_threshold:
            int_array, scale_list, dim_list, total_dim = self.eden.compress(data, seed)

            # TODO: workaround: using the int to float dictionary to pass eden's metadata
            metadata["int_to_float"] = {0: float(seed), 1: float(total_dim)}

            k = 2
            for scale, dim in zip(scale_list, dim_list):
                metadata["int_to_float"][k] = scale
                metadata["int_to_float"][k + 1] = float(dim)
                k += 2

            return_values = int_array.astype(np.uint8).tobytes(), metadata

        else:
            no_comp_data, metadata = self.no_comp.forward(data)
            return_values = no_comp_data, metadata

        return return_values

    def backward(self, data, metadata, **kwargs):
        """Recover data array back to the original numerical type and the
        shape.

        Args:
            data: an flattened numpy array.
            metadata: dictionary to contain information for recovering to
                original data array.

        Returns:
            data: Numpy array with original numerical type and shape.
        """

        if np.prod(metadata["int_list"]) >= self.dim_threshold:  # compressed data
            data = np.frombuffer(data, dtype=np.uint8)
            data = co.deepcopy(data)
            data = self.eden.decompress(data, metadata["int_to_float"])
            data_shape = list(metadata["int_list"])
            data = data.reshape(data_shape)
        else:
            data = self.no_comp.backward(data, metadata)
        data = data.astype(np.float32)

        return data


class EdenPipeline(TransformationPipeline):
    """A pipeline class for compressing data using the Eden method.

    This class is a pipeline of transformers that use the Eden method for
    quantization.

    Attributes:
        n_bits (int): The number of bits per coordinate for quantization.
        dim_threshold (int): The threshold for the dimension of the data. Data
            with dimensions less than this threshold are not compressed.
        device (str): The device to be used for quantization ('cpu' or 'cuda').
    """

    def __init__(self, n_bits=8, dim_threshold=100, device="cpu", **kwargs):
        """Initialize a pipeline of transformers.

        Args:
            n_bits (int): Number of bits per coordinate (1-8 bits are
                supported).
            dim_threshold (int): Layers with less than dim_threshold params
                are not compressed.
            device: Device for executing the compression and decompressionc
                (e.g., 'cpu', 'cuda:0', 'cuda:1').

        Return:
            Transformer class object.
        """

        # instantiate each transformer
        transformers = [EdenTransformer(n_bits, dim_threshold, device)]
        super().__init__(transformers=transformers, **kwargs)
